from time import time
from typing import List
from tdw.add_ons.add_on import AddOn


class Benchmark(AddOn):
    """
    Benchmark the frames per second (FPS) over a given number of frames.
    """

    def __init__(self):
        """
        (no parameters)
        """

        super().__init__()
        self.initialized = True
        """:field
        A list of time elapsed per `communicate()` call.
        """
        self.times: List[float] = list()
        # If True, we are currently benchmarking.
        self._benchmarking: bool = False
        # The initial time.
        self._t0: float = -1
        """:field
        The frames per second of the previous benchmark test.
        """
        self.fps: float = -1

    def get_initialization_commands(self) -> List[dict]:
        return []

    def before_send(self, commands: List[dict]) -> None:
        if self._benchmarking:
            self._t0 = time()

    def on_send(self, resp: List[bytes]) -> None:
        # Clock the benchmark.
        if self._benchmarking:
            t1 = time()
            self.times.append(t1 - self._t0)
            self._t0 = t1

    def start(self) -> None:
        """
        Start bencharking each `communicate()` call and clear `self.times`.
        """

        self.fps = -1
        self.times.clear()
        self._benchmarking = True

    def stop(self) -> None:
        """
        Stop benchmarking each `communicate()` call and set `self.fps`.
        """

        self._benchmarking = False
        self.fps = len(self.times) / sum(self.times)



from typing import List, Union
from pathlib import Path
import numpy as np
from tdw.output_data import OutputData, AudioSources, Rigidbodies, Transforms
from tdw.audio_utils import AudioUtils
from tdw.add_ons.add_on import AddOn


class PhysicsAudioRecorder(AddOn):
    """
    Record audio generated by physics events.
    """

    def __init__(self, max_frames: int = -1):
        """
        :param max_frames: If greater than 0, stop recording after this many frames even if objects are still moving or making sound.
        """

        super().__init__()

        """:field
        # If greater than 0, stop recording after this many frames even if objects are still moving or making sound.
        """
        self.max_frames: int = max_frames
        # The current frame.
        self._frame: int = 0
        """:field
        The path to the next audio file.
        """
        self.path: Path = Path.home()
        """:field
        If True, there is an ongoing recording.
        """
        self.recording: bool = False

    def get_initialization_commands(self) -> List[dict]:
        return []

    def on_send(self, resp: List[bytes]) -> None:
        if not self.recording:
            return
        # Stop recording at the maximum number of frames.
        self._frame += 1
        if 0 < self.max_frames <= self._frame:
            AudioUtils.stop()
            return
        # Get any objects that fell below the floor.
        below_floor: List[int] = list()
        for i in range(len(resp) - 1):
            r_id = OutputData.get_data_type_id(resp[i])
            if r_id == "tran":
                transforms = Transforms(resp[i])
                for j in range(transforms.get_num()):
                    if transforms.get_position(j)[1] < -0.1:
                        below_floor.append(transforms.get_id(j))
        # Check if objects have stopped moving and no audio is playing.
        sleeping = True
        playing_audio = False
        for i in range(len(resp) - 1):
            r_id = OutputData.get_data_type_id(resp[i])
            if r_id == "rigi":
                rigidbodies = Rigidbodies(resp[i])
                for j in range(rigidbodies.get_num()):
                    if rigidbodies.get_id(j) not in below_floor and not rigidbodies.get_sleeping(j):
                        sleeping = False
                        break
            elif r_id == "audi":
                audio_sources = AudioSources(resp[i])
                for j in range(audio_sources.get_num()):
                    if audio_sources.get_is_playing(j):
                        playing_audio = True
                        break
                # Check if the simulation is totally silent (there might be Resonance Audio reverb).
                if not playing_audio and np.max(audio_sources.get_samples()) > 0:
                    playing_audio = True
        if sleeping and not playing_audio:
            self.stop()

    def start(self, path: Union[str, Path]) -> None:
        """
        Start recording.

        :param path: The path to the output .wav file.
        """

        # Don't start a new recording if one is ongoing.
        if self.recording:
            return
        self.recording = True
        if isinstance(path, str):
            self.path = Path(path)
        else:
            self.path = path
        if not self.path.parent.exists:
            self.path.parent.mkdir(parents=True)
        if self.path.exists():
            self.path.unlink()

        self._frame = 0
        # Start listening.
        AudioUtils.start(output_path=self.path)
        self.commands.extend([{"$type": "send_audio_sources",
                               "frequency": "always"},
                              {"$type": "send_rigidbodies",
                               "frequency": "always"},
                              {"$type": "send_transforms",
                               "frequency": "always"}])

    def stop(self) -> None:
        """
        Stop an ongoing recording. Use ffmpeg to remove initial silence.
        """

        AudioUtils.stop()
        self.recording = False



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FBOutput

import tdw.flatbuffers

class Category(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsCategory(cls, buf, offset):
        n = tdw.flatbuffers.encode.Get(tdw.flatbuffers.packer.uoffset, buf, offset)
        x = Category()
        x.Init(buf, n + offset)
        return x

    # Category
    def Init(self, buf, pos):
        self._tab = tdw.flatbuffers.table.Table(buf, pos)

    # Category
    def Color(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from .Color import Color
            obj = Color()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Category
    def Name(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

def CategoryStart(builder): builder.StartObject(2)
def CategoryAddColor(builder, color): builder.PrependStructSlot(0, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(color), 0)
def CategoryAddName(builder, name): builder.PrependUOffsetTRelativeSlot(1, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)
def CategoryEnd(builder): return builder.EndObject()



class ScrapeSubObject:
    """
    Data for a sub-object of a model being used as a scrape surface.
    """

    def __init__(self, name: str, material_index: int):
        """
        :param name: The name of the sub-object.
        :param material_index: The index of the material.
        """

        """:field
        The name of the sub-object.
        """
        self.name: str = name
        """:field
        The index of the material.
        """
        self.material_index: int = material_index



from pathlib import Path
from typing import List, Union
from json import load, dump
from tdw.output_data import OutputData, LogMessage
from tdw.add_ons.add_on import AddOn


class Logger(AddOn):
    """
    Record and playback every command sent to the build.

    ```python
    from tdw.controller import Controller
    from tdw.add_ons.logger import Logger

    c = Controller()
    logger = Logger(record=True, path="log.json")
    c.add_ons.append(logger)
    # The logger add-on will log this command.
    c.communicate({"$type": "do_nothing"})
    # The logger add-on will log this command and generate a log.json file.
    c.communicate({"$type": "terminate"})
    ```
    """

    def __init__(self, record: bool, path: Union[str, Path], log_commands_in_build: bool = False):
        """
        :param record: If True, record each command. If False, play back an existing record.
        :param path: The path to either save the record to or load the record from.
        :param log_commands_in_build: If True, the build will log every message received and every command executed in the [Player log](https://docs.unity3d.com/Manual/LogFiles.html).
        """

        super().__init__()

        # If True, the build will log every message received and every command executed in the Player log.
        self._log_commands_in_build: bool = log_commands_in_build

        """:field
        If True, record each command. If False, play back an existing record.
        """
        self.record: bool = record

        # Get or create the playback file path.
        if isinstance(path, str):
            self._path: Path = Path(path)
        else:
            self._path: Path = path
        if not self._path.parent.exists:
            self._path.parent.mkdir(parents=True)

        # Start a new playback file.
        if self.record:
            """:field
            A record of each list of commands sent to the build.
            """
            self.playback: List[List[dict]] = list()
        # Load an existing .json file.
        else:
            with self._path.open("rt", encoding="utf-8") as f:
                self.playback = load(f)

    def on_send(self, resp: List[bytes]) -> None:
        # Prepare to send the next list of commands.
        if not self.record:
            if len(self.playback) > 0:
                self.commands = self.playback.pop(0)
        for i in range(len(resp) - 1):
            r_id = OutputData.get_data_type_id(resp[i])
            # Print a log message.
            if r_id == "logm":
                log = LogMessage(resp[i])
                print(f"[FROM BUILD] {log.get_message_type()} from {log.get_object_type()}: {log.get_message()}")
            # If we get a quit signal and we're recording, save the log file.
            elif r_id == "quit" and self.record:
                self.save()

    def get_initialization_commands(self) -> List[dict]:
        commands = [{"$type": "send_log_messages"}]
        if self._log_commands_in_build:
            commands.append({"$type": "set_network_logging",
                             "value": True})
        return commands

    def before_send(self, commands: List[dict]) -> None:
        # Record the commands that were just sent.
        if self.record:
            self.playback.append(commands[:])

    def save(self) -> None:
        """
        Write the record of commands sent to the local disk.
        """

        with self._path.open("wt", encoding="utf-8") as f:
            dump(self.playback, f)



from typing import Dict, List
from tdw.output_data import OutputData, Collision, EnvironmentCollision
from tdw.collision_data.collision_obj_obj import CollisionObjObj
from tdw.collision_data.collision_obj_env import CollisionObjEnv
from tdw.int_pair import IntPair
from tdw.add_ons.add_on import AddOn


class CollisionManager(AddOn):
    """
    Manager add-on for all collisions on this frame.
    """
    
    def __init__(self, enter: bool = True, stay: bool = False, exit: bool = False,
                 objects: bool = True, environment: bool = True):
        """
        :param enter: If True, listen for collision enter events.
        :param stay: If True, listen for collision stay events.
        :param exit: If True, listen for collision exit events.
        :param objects: If True, listen for collisions between objects.
        :param environment: If True, listen for collisions between an object and the environment.
        """

        super().__init__()
        collision_types: List[str] = list()
        if objects:
            collision_types.append("obj")
        if environment:
            collision_types.append("env")
        self._send_collision_commands: dict = {"$type": "send_collisions",
                                               "enter": enter,
                                               "stay": stay,
                                               "exit": exit,
                                               "collision_types": collision_types}
        """:field
        All collisions between two objects that occurred on the frame.
        Key = An `IntPair` (a pair of object IDs). Value = [The collision.](../collision_data/collision_obj_obj.md)
        """
        self.obj_collisions: Dict[IntPair, CollisionObjObj] = dict()
        """:field
        All collisions between an object and the environment that occurred on the frame.
        Key = the object ID. Value = [The collision.](../collision_data/collision_obj_env.md)
        """
        self.env_collisions: Dict[int, CollisionObjEnv] = dict()

    def get_initialization_commands(self) -> List[dict]:
        return [self._send_collision_commands]

    def on_send(self, resp: List[bytes]) -> None:
        self.obj_collisions.clear()
        self.env_collisions.clear()
        for i in range(len(resp) - 1):
            r_id = OutputData.get_data_type_id(resp[i])
            if r_id == "coll":
                collision = Collision(resp[i])
                # Get the pair of IDs in this collision and use it as a key.
                ids = IntPair(int1=collision.get_collider_id(), int2=collision.get_collidee_id())
                coo = CollisionObjObj(collision=collision)
                self.obj_collisions[ids] = coo
            elif r_id == "enco":
                collision = EnvironmentCollision(resp[i])
                coe = CollisionObjEnv(collision=collision)
                self.env_collisions[collision.get_object_id()] = coe



from typing import List
from abc import ABC, abstractmethod


class AddOn(ABC):
    """
    Controller add-ons can be "attached" to any controller to add functionality into the `communicate()` function.

    Add-ons work by reading the response from the build and building a list of commands to be sent on the next frame.
    Anything that add-ons do can be replicated elsewhere via the TDW Command API, which means that these add-ons don't provide _additional_ functionality to TDW; rather, they are utility objects for commonly required tasks such as image capture.

    We recommend that new TDW users use add-ons in their controllers, while more experienced users might prefer to have more fine-grained control. Add-ons are a new feature in TDW as of v1.9.0 and we're still in the process of updating our example controllers.

    ## Usage

    To attach an add-on, append it to the `add_ons` list.
    Every time `communicate()` is called, the add-on will evaluate the response from the build. The add-on can send additional commands to the build on the next frame or do something within its own state (such as update an ongoing log):

    ```python
    from tdw.controller import Controller
    from tdw.add_ons.logger import Logger

    c = Controller(launch_build=False)
    logger = Logger(record=True, path="log.json")
    c.add_ons.append(logger)
    # The logger add-on will log this command.
    c.communicate({"$type": "do_nothing"})
    # The logger add-on will log this command and generate a log.json file.
    c.communicate({"$type": "terminate"})
    ```
    """

    def __init__(self):
        """
        (no parameters)
        """

        """:field
        These commands will be appended to the commands of the next `communicate()` call.
        """
        self.commands: List[dict] = list()
        """:field
        If True, this module has been initialized.
        """
        self.initialized: bool = False

    @abstractmethod
    def get_initialization_commands(self) -> List[dict]:
        """
        This function gets called exactly once per add-on. To re-initialize, set `self.initialized = False`.

        :return: A list of commands that will initialize this add-on.
        """

        raise Exception()

    @abstractmethod
    def on_send(self, resp: List[bytes]) -> None:
        """
        This is called after commands are sent to the build and a response is received.

        Use this function to send commands to the build on the next frame, given the `resp` response.
        Any commands in the `self.commands` list will be sent on the next frame.

        :param resp: The response from the build.
        """

        raise Exception()

    def before_send(self, commands: List[dict]) -> None:
        """
        This is called before sending commands to the build. By default, this function doesn't do anything.

        :param commands: The commands that are about to be sent to the build.
        """

        pass



from enum import Enum


class AudioMaterial(Enum):
    """
    These are the materials currently supported for impact sounds in [`PyImpact`](../add_ons/py_impact.md). More will be added in time.
    """

    ceramic = 0
    glass = 1
    metal = 2
    wood_hard = 3
    wood_medium = 4
    wood_soft = 5
    cardboard = 6
    paper = 7
    plastic_hard = 8
    plastic_soft_foam = 9
    rubber = 10
    fabric = 11
    leather = 12
    stone = 13



import io
from csv import DictReader
from pathlib import Path
from pkg_resources import resource_filename
from typing import Union, Dict
from tdw.physics_audio.audio_material import AudioMaterial


class ObjectAudioStatic:
    """
    Impact sound data for an object in a TDW model library.
    The audio values here are just recommendations; you can apply different values if you want.
    """

    def __init__(self, name: str, amp: float, mass: float, material: AudioMaterial, bounciness: float, resonance: float, size: int, object_id: int):
        """
        :param name: The model name.
        :param amp: The sound amplitude.
        :param mass: The object mass.
        :param material: The audio material.
        :param bounciness: The bounciness value for a Unity physics material.
        :param resonance: The resonance value for the object.
        :param size: Integer representing the size "bucket" this object belongs to (0-5).
        :param object_id: The ID of the object.
        """

        """:field
        The sound amplitude.
        """
        self.amp: float = amp
        """:float
        The object mass.
        """
        self.mass: float = mass
        """:field
        The audio material.
        """
        self.material: AudioMaterial = material
        """:field
        The name of the object.
        """
        self.name: str = name
        """:field
        The bounciness value for a Unity physics material. 
        """
        self.bounciness: float = bounciness
        """:field
        The resonance value for the object.
        """
        self.resonance: float = resonance
        """:field
        Integer representing the size "bucket" this object belongs to (0-5).
        """
        self.size = size
        """:field
        The ID of the object.
        """
        self.object_id = object_id


def get_static_audio_data(csv_file: Union[str, Path] = "") -> Dict[str, ObjectAudioStatic]:
    """
    Returns ObjectInfo values.
    As of right now, only a few objects in the TDW model libraries are included. More will be added in time.

    :param csv_file: The path to the .csv file containing the object info. By default, it will load `tdw/physics_audio/objects.csv`. If you want to make your own spreadsheet, use this file as a reference.

    :return: A list of default ObjectInfo. Key = the name of the model. Value = object info.
    """

    objects: Dict[str, ObjectAudioStatic] = {}
    # Load the objects.csv metadata file.
    if isinstance(csv_file, str):
        # Load the default file.
        if csv_file == "":
            csv_file = str(Path(resource_filename(__name__, f"objects.csv")).resolve())
        else:
            csv_file = str(Path(csv_file).resolve())
    else:
        csv_file = str(csv_file.resolve())

    # Parse the .csv file.
    with io.open(csv_file, newline='', encoding='utf-8-sig') as f:
        reader = DictReader(f)
        for row in reader:
            o = ObjectAudioStatic(name=row["name"], amp=float(row["amp"]), mass=float(row["mass"]),
                                  material=AudioMaterial[row["material"]], bounciness=float(row["bounciness"]),
                                  resonance=float(row["resonance"]), size=int(row["size"]), object_id=0)
            objects.update({o.name: o})
    return objects


DEFAULT_OBJECT_AUDIO_STATIC_DATA: Dict[str, ObjectAudioStatic] = get_static_audio_data()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FBOutput

import tdw.flatbuffers

class PointLight(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsPointLight(cls, buf, offset):
        n = tdw.flatbuffers.encode.Get(tdw.flatbuffers.packer.uoffset, buf, offset)
        x = PointLight()
        x.Init(buf, n + offset)
        return x

    # PointLight
    def Init(self, buf, pos):
        self._tab = tdw.flatbuffers.table.Table(buf, pos)

    # PointLight
    def Intensity(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(tdw.flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # PointLight
    def Color(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = o + self._tab.Pos
            from .Color import Color
            obj = Color()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # PointLight
    def Position(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # PointLight
    def Range(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(tdw.flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

def PointLightStart(builder): builder.StartObject(4)
def PointLightAddIntensity(builder, intensity): builder.PrependFloat32Slot(0, intensity, 0.0)
def PointLightAddColor(builder, color): builder.PrependStructSlot(1, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(color), 0)
def PointLightAddPosition(builder, position): builder.PrependStructSlot(2, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(position), 0)
def PointLightAddRange(builder, range): builder.PrependFloat32Slot(3, range, 0.0)
def PointLightEnd(builder): return builder.EndObject()



from typing import List
from io import BytesIO
from PIL import Image
from collections import Counter
from tdw.add_ons.model_verifier.model_tests.rotate_object_test import RotateObjectTest
from tdw.output_data import Images
from tdw.librarian import ModelRecord


class PhysicsQuality(RotateObjectTest):
    """
    Test the "physics quality" i.e. the disparity between the colliders volume and the rendered volume.
    """

    MASK_RED = (255, 0, 0)

    def __init__(self, record: ModelRecord):
        super().__init__(record=record)
        self._showing_collider_hulls: bool = False
        self._without_collider_hulls: List[int] = list()
        self._with_collider_hulls: List[int] = list()

    def start(self) -> List[dict]:
        """
        :return: A list of commands to start the test.
        """

        commands = super().start()
        commands.insert(0, {"$type": "set_pass_masks",
                            "pass_masks": ["_mask"]})
        return commands

    def on_send(self, resp: List[bytes]) -> List[dict]:
        """
        :param resp: The response from the build.

        :return: A list of commands to continue or end the test.
        """

        commands = super().on_send(resp=resp)
        if self.done:
            # Done with both passes. Get the physics quality.
            if self._showing_collider_hulls:
                qualities: List[float] = list()
                for with_colliders, without_colliders in zip(self._with_collider_hulls, self._without_collider_hulls):
                    if without_colliders == 0:
                        physics_quality = 0
                    else:
                        physics_quality = 1 - float(without_colliders - with_colliders) / without_colliders
                    qualities.append(physics_quality)
                if len(qualities) == 0:
                    physics_quality = -1
                else:
                    physics_quality = float(sum(qualities)) / len(qualities)
                self.reports.append(str(physics_quality))
                commands.extend(RotateObjectTest._get_end_commands())
                return commands
            # Show the colliders and start rotating again.
            else:
                self._showing_collider_hulls = True
                self.done = False
                self._angle = 0
                self._axis = "yaw"
                return [{"$type": "show_collider_hulls",
                         "id": RotateObjectTest.OBJECT_ID},
                        {"$type": "set_pass_masks",
                         "pass_masks": ["_img"]},
                        {"$type": "rotate_object_to",
                         "rotation": {"w": 1, "x": 0, "y": 0, "z": 0},
                         "id": RotateObjectTest.OBJECT_ID},
                        {"$type": "teleport_object",
                         "id": RotateObjectTest.OBJECT_ID,
                         "position": {"x": 0, "y": 0.5, "z": 0}}]
        else:
            return commands

    def _read_images(self, images: Images) -> None:
        image = Image.open(BytesIO(images.get_image(0)))
        if self._showing_collider_hulls:
            self._with_collider_hulls.append(Counter(image.getdata())[RotateObjectTest.PINK])
        else:
            self._without_collider_hulls.append(Counter(image.getdata())[PhysicsQuality.MASK_RED])



from typing import List, Union, Dict
from pathlib import Path
from PIL.Image import Image
from tdw.add_ons.add_on import AddOn
from tdw.tdw_utils import TDWUtils
from tdw.output_data import OutputData, Images


class ImageCapture(AddOn):
    """
    Request image data and save the images to disk. By default, images will be saved every frame for each specified avatar, but this add-on can be reset to save images for specific avatars, certain frames, etc.

    Note that image capture in TDW is *not* the same as image *rendering*. An avatar can render image to a display without actually sending them to the controller.
    Sending images and image passes (such as the `_id` segmentation color pass) is the slowest process in TDW; only request image capture when you actually need it.

    ```python
    from tdw.controller import Controller
    from tdw.tdw_utils import TDWUtils
    from tdw.add_ons.third_person_camera import ThirdPersonCamera
    from tdw.add_ons.image_capture import ImageCapture

    c = Controller(launch_build=False)

    # Add a third-person camera. It will look at object 0.
    object_id = 0
    camera = ThirdPersonCamera(position={"x": 0.5, "y": 1.5, "z": -2},
                               look_at=object_id)
    # Tell the camera to capture images per-frame.
    capture = ImageCapture(avatar_ids=[camera.avatar_id], path="D:/image_capture_test", pass_masks=["_img", "_id"])
    c.add_ons.extend([camera, capture])

    # Create an empty room and add an object.
    # The camera will be added after creating the empty room and the object.
    # The image capture add-on will initialize after the camera and save an `_img` pass and `_id` pass to disk.
    c.communicate([TDWUtils.create_empty_room(12, 12),
                   c.get_add_object(model_name="iron_box",
                                    object_id=object_id)])

    c.communicate({"$type": "terminate"})
    ```
    """

    # A list of valid pass masks.
    _PASS_MASKS: List[str] = list(Images.PASS_MASKS.values())

    def __init__(self, path: Union[str, Path], avatar_ids: List[str] = None, png: bool = False, pass_masks: List[str] = None):
        """
        :param path: The path to the output directory.
        :param avatar_ids: The IDs of the avatars that will capture and save images. If empty, all avatars will capture and save images. Note that these avatars must already exist in the scene (if you've added the avatars via a [`ThirdPersonCamera` add-on](third_person_camera.md), you must add the `ThirdPersonCamera` first, *then* `ImageCapture`).
        :param png: If True, images will be lossless png files. If False, images will be jpgs. Usually, jpg is sufficient.
        :param pass_masks: A list of image passes that will be captured by the avatars. If None, defaults to `["_img"]`. For a description of each of pass mask, [read this](https://github.com/threedworld-mit/tdw/blob/master/Documentation/api/command_api.md#set_pass_masks).
        """

        super().__init__()
        """:field
        The current frame count. This is used to generate filenames.
        """
        self.frame: int = 0
        if isinstance(path, str):
            """:field
            The path to the output directory.
            """
            self.path: Path = Path(path)
        else:
            self.path: Path = path
        if not self.path.exists():
            self.path.mkdir(parents=True)
        if avatar_ids is None:
            """:field
            The IDs of the avatars that will capture and save images. If empty, all avatars will capture and save images.
            """
            self.avatar_ids: List[str] = []
        else:
            self.avatar_ids: List[str] = avatar_ids
        # If True, encode the _img pass as a png.
        self._png: bool = png
        # A list of pass mask commands to send on the next frame.
        self._pass_mask_commands: List[dict] = self._get_pass_mask_commands(pass_masks=pass_masks)
        # The frequency at which images are sent.
        self._frequency: str = "always"
        # If True, save images per frame.
        self._save: bool = True

        """:field
        Raw [`Images` output data](../../api/output_data.md#Images) from the build. Key = The ID of the avatar. This is updated per frame. If an avatar didn't capture an image on this frame, it won't be in this dictionary.
        """
        self.images: Dict[str, Images] = dict()

    def get_initialization_commands(self) -> List[dict]:
        commands = [{"$type": "set_img_pass_encoding",
                     "value": self._png}]
        # Get the pass masks.
        commands.extend(self._pass_mask_commands)
        # Begin by sending images for the next frame.
        commands.append({"$type": "send_images",
                         "frequency": "once",
                         "ids": self.avatar_ids})
        return commands

    def on_send(self, resp: List[bytes]) -> None:
        got_images = False
        self.images.clear()
        for i in range(len(resp) - 1):
            r_id = OutputData.get_data_type_id(resp[i])
            if r_id == "imag":
                images = Images(resp[i])
                a = images.get_avatar_id()
                # Store the image data.
                self.images[a] = images
                if self._save and (len(self.avatar_ids) == 0 or a in self.avatar_ids):
                    output_dir = self.path.joinpath(a)
                    if not output_dir.exists():
                        output_dir.mkdir(parents=True)
                    # Save images.
                    TDWUtils.save_images(images=images,
                                         output_directory=str(output_dir.resolve()),
                                         filename=TDWUtils.zero_padding(self.frame, 4))
                    got_images = True
        if got_images:
            self.frame += 1
        # If we're requesting images per-frame, send the command.
        # We can't use the "always" value because of cases like that Magnebot that will turn off image capture.
        if self._frequency == "always":
            self.commands.append({"$type": "send_images",
                                  "frequency": "once",
                                  "ids": self.avatar_ids})

    def set(self, frequency: str = "always", avatar_ids: List[str] = None, pass_masks: List[str] = None, save: bool = True) -> None:
        """
        Set the frequency of images and which avatars will capture images.
        By default, all of the avatars specified in the constructor (if None, all avatars in the scene) will capture images every frame.
        This function will override the previous image capture settings; in other words, setting `frequency` to `"once"` for one avatar will make all other avatars stop capturing images per frame.

        :param frequency: The frequency at which images are captured. Options: `"always"` (capture images every frame), `"once"` (capture an image only on the next frame), `"never"` (stop capturing images).
        :param avatar_ids: The IDs of the avatar that will capture images. If None, all avatars in the scene will capture images.
        :param pass_masks: A list of image passes that will be captured by the avatars. If None, defaults to `["_img"]`. For a description of each of pass mask, [read this](https://github.com/threedworld-mit/tdw/blob/master/Documentation/api/command_api.md#set_pass_masks).
        :param save: If True, automatically save images to disk per frame. If False, images won't be saved but the `self.images` dictionary will still be updated.
        """

        self.avatar_ids: List[str] = avatar_ids if avatar_ids is not None else []
        self._pass_mask_commands = self._get_pass_mask_commands(pass_masks=pass_masks)
        self.commands.extend(self._pass_mask_commands)
        self._save = save
        self._frequency = frequency
        if self._frequency == "always":
            # This is handled in on_send()
            pass
        # Send images once or never.
        elif self._frequency == "once" or self._frequency == "never":
            self.commands.append({"$type": "send_images",
                                  "frequency": self._frequency,
                                  "ids": self.avatar_ids})
        else:
            raise Exception(f"Invalid frequency: {self._frequency}")

    def get_pil_images(self) -> Dict[str, Dict[str, Image]]:
        """
        Convert the latest image data from the build (`self.images`) to PIL images. Note that it is not necessary to call this function to save images; use this only to analyze an image at runtime.

        :return: A dictionary of PIL images from the latest image data from the build. Key = The avatar ID. Value = A dictionary; key = the pass mask, value = the PIL image.
        """

        images: Dict[str, Dict[str, Image]] = dict()
        for avatar_id in self.images:
            images[avatar_id] = dict()
            for i in range(self.images[avatar_id].get_num_passes()):
                images[avatar_id][self.images[avatar_id].get_pass_mask(i)] = \
                    TDWUtils.get_pil_image(images=self.images[avatar_id], index=i)
        return images

    def _get_pass_mask_commands(self, pass_masks: List[str] = None) -> List[dict]:
        """
        :param pass_masks: The pass masks. If None, defaults to `["_img"]`.

        :return: A list of commands to set pass masks per avatar.
        """

        # Set the pass masks.
        if pass_masks is None:
            pass_masks = ["_img"]
        for pm in pass_masks:
            if pm not in ImageCapture._PASS_MASKS:
                raise Exception(f"Invalid pass mask: {pm}")
        return [{"$type": "set_pass_masks", "pass_masks": pass_masks, "avatar_id": a} for a in self.avatar_ids]




from typing import List, Dict, Optional, Union
from os import getcwd, chdir, walk, devnull
import re
from platform import system
from pathlib import Path
from subprocess import call
from distutils.file_util import copy_file, move_file
from distutils.dir_util import remove_tree
from tdw.asset_bundle_creator_base import AssetBundleCreatorBase
from tdw.librarian import RobotRecord
from tdw.backend.paths import EDITOR_LOG_PATH
from tdw.backend.platforms import UNITY_TO_SYSTEM


class RobotCreator(AssetBundleCreatorBase):
    """
    Download a .urdf or .xacro file and convert it into an asset bundle that is usable by TDW.
    """

    # The root temporary directory.
    TEMP_ROOT = Path.home().joinpath("robot_creator/temp_robots")

    def __init__(self, quiet: bool = False, display: str = ":0", unity_editor_path: Union[Path, str] = None):
        """
        :param quiet: If true, don't print any messages to console.
        :param display: The display to launch Unity Editor on. Ignored if this isn't Linux.
        :param unity_editor_path: The path to the Unity Editor executable, for example `C:/Program Files/Unity/Hub/Editor/2020.3.24f1/Editor/Unity.exe`. If None, this script will try to find Unity Editor automatically.
        """

        super().__init__(quiet=quiet, display=display, unity_editor_path=unity_editor_path)

    def create_asset_bundles(self, urdf_url: str, required_repo_urls: Dict[str, str] = None,
                             xacro_args: Dict[str, str] = None, immovable: bool = True, up: str = "y",
                             description_infix: str = None, branch: str = None) -> RobotRecord:
        """
        Given the URL of a .urdf file or a .xacro file, create asset bundles of the robot.

        This is a wrapper function for:

        1. `clone_repo()`
        2. `copy_files()`
        3. `urdf_to_prefab()`
        4. `prefab_to_asset_bundles()`

        :param urdf_url: The URL of a .urdf or a .xacro file.
        :param required_repo_urls: A dictionary of description folder names and repo URLs outside of the robot's repo that are required to create the robot. This is only required for .xacro files that reference outside repos. For example, the Sawyer robot requires this to add the gripper: `{"intera_tools_description": "https://github.com/RethinkRobotics/intera_common"}`
        :param xacro_args: Names and values for the `arg` tags in the .xacro file (ignored if this is a .urdf file). For example, the Sawyer robot requires this to add the gripper: `{"electric_gripper": "true"}`
        :param immovable: If True, the base of the robot is immovable.
        :param up: The up direction. Used when importing the robot into Unity. Options: `"y"` or `"z"`. Usually, this should be the default value (`"y"`).
        :param description_infix: The name of the description infix within the .urdf URL, such as `fetch_description`. Only set this if the urdf URL is non-standard; otherwise `RobotCreator` should be able to find this automatically.
        :param branch: The name of the branch of the repo. If None, defaults to `"master"`.

        :return: A `RobotRecord` object. The `urls` field contains the paths to each asset bundle.
        """

        if required_repo_urls is None:
            required_repo_urls = list()

        # Clone the repo.
        repo_paths: Dict[str, Path] = dict()
        local_repo_path = self.clone_repo(url=urdf_url)
        if description_infix is None:
            description_infix = RobotCreator._get_description_infix(url=urdf_url)
        repo_paths[description_infix] = local_repo_path

        # Clone the required repos.
        for description in required_repo_urls:
            required_repo_url = required_repo_urls[description]
            required_local_repo_path = self.clone_repo(url=required_repo_url)
            repo_paths[description] = required_local_repo_path

        # Copy the files, create a .urdf file (if needed), and creator collider objects.
        urdf_path = self.copy_files(urdf_url=urdf_url, local_repo_path=local_repo_path, repo_paths=repo_paths,
                                    xacro_args=xacro_args, branch=branch)

        # Create the prefab.
        prefab = self.urdf_to_prefab(urdf_path=urdf_path, immovable=immovable, up=up)
        name = prefab.name.replace(".prefab", "")

        # Create the asset bundles.
        asset_bundles = self.prefab_to_asset_bundles(name=name)
        temp = dict()
        for k in asset_bundles:
            temp[k] = "file:///" + str(asset_bundles[k].resolve()).replace("\\", "/")
        asset_bundles = temp

        # Create the record.
        return RobotCreator.get_record(name=name, urdf_url=urdf_url, immovable=immovable, asset_bundles=asset_bundles)

    @staticmethod
    def get_record(name: str, urdf_url: str, immovable: bool, asset_bundles: Dict[str, str]) -> RobotRecord:
        """
        :param name: The name of the robot.
        :param urdf_url: The URL to the .urdf or .xacro file.
        :param immovable: If True, the base of the robot is immovable.
        :param asset_bundles: The paths to the asset bundles. See `prefab_to_asset_bundles()`.

        :return: A `RobotRecord` metadata object.
        """

        record_data = {"name": name,
                       "source": RobotCreator._get_repo_url(url=urdf_url),
                       "immovable": immovable,
                       "urls": asset_bundles,
                       "targets": {},
                       "ik": {}}
        return RobotRecord(data=record_data)

    def clone_repo(self, url: str) -> Path:
        """
        Clone a repo to a temporary directory.

        :param url: The URL to the .urdf or .xacro file or the repo.

        :return: The temporary directory.
        """

        if not RobotCreator.TEMP_ROOT.exists():
            RobotCreator.TEMP_ROOT.mkdir(parents=True)

        # This is a .urdf or .xacro file. Parse the repo URL accordingly.
        if url.endswith(".xacro") or url.endswith(".urdf"):
            local_repo_path = RobotCreator._get_local_repo_path(url=url)
            repo_url = RobotCreator._get_repo_url(url=url)
        # This is the base URL of the repo. Parse it accordingly.
        else:
            local_repo_path = RobotCreator.TEMP_ROOT.joinpath(Path(url).name)
            repo_url = url
        if local_repo_path.exists():
            return local_repo_path

        # Change directory.
        cwd = getcwd()
        chdir(str(RobotCreator.TEMP_ROOT.resolve()))
        if not self._quiet:
            print(f"Cloning: {repo_url}")
        # Clone the repo.
        call(["git", "clone", repo_url],
             stderr=open(devnull, "wb"))
        chdir(cwd)
        assert local_repo_path.exists(), f"Can't find: {local_repo_path.resolve()}"
        if not self._quiet:
            print("...Done!")
        return local_repo_path

    def copy_files(self, urdf_url: str, local_repo_path: Path, repo_paths: Dict[str, Path],
                   xacro_args: Dict[str, str] = None, branch: str = None) -> Path:
        """
        Copy and convert files required to create a prefab.

        1. If this is a .xacro file, convert it to a .urdf file.
        2. Copy the .urdf file to the Unity project.
        3. Copy all associated meshes to the .urdf project.

        :param urdf_url: The URL to the remote .urdf or .xacro file.
        :param local_repo_path: The path to the local repo.
        :param repo_paths: A dictionary of required repos (including the one that the .urdf or .xacro is in). Key = The description path infix, e.g. "sawyer_description". Value = The path to the local repo.
        :param xacro_args: Names and values for the `arg` tags in the .xacro file. Can be None for a .urdf or .xacro file and always ignored for a .urdf file.
        :param branch: The name of the branch of the repo. If None, defaults to `"master"`.

        :return: The path to the .urdf file in the Unity project.
        """

        page_url = self._raw_to_page(url=urdf_url)
        if branch is None:
            branch = "master"
        repo_path = re.search(r"(.*)/blob/" + branch + r"/(.*)", page_url).group(2)
        urdf_path = local_repo_path.joinpath(repo_path)
        dst_root = self._project_path.joinpath(f"Assets/robots")
        if not dst_root.exists():
            dst_root.mkdir(parents=True)
        # Convert the .xacro file to a .urdf file.
        if Path(urdf_url).suffix == ".xacro":
            urdf_dst = self.xacro_to_urdf(xacro_path=urdf_path, repo_paths=repo_paths, args=xacro_args)
        # Move the existing .urdf file.
        else:
            assert urdf_path.exists(), f"Not found: {urdf_path.resolve()}"
            # Copy the .urdf file.
            urdf_dst = dst_root.joinpath(Path(urdf_url).name)
            copy_file(src=str(urdf_path.resolve()), dst=str(urdf_dst.resolve()))

        # Read the .urdf file.
        urdf = urdf_dst.read_text(encoding="utf-8")
        # Remove gazebo stuff.
        urdf = re.sub(r"<gazebo reference(.*?)>((.|\n)*?)</gazebo>", "", urdf)
        urdf = re.sub(r"<gazebo>((.|\n)*?)</gazebo>", "", urdf)
        urdf = re.sub(r"<transmission((.|\n)*?)</transmission>", "", urdf)
        urdf = re.sub(r'<xacro:include (.*)gazebo(.*)/>', "", urdf)
        urdf_dst.write_text(urdf, encoding="utf-8")
        # Copy the meshes.
        for m in re.findall(r"filename=\"package://((.*)\.(DAE|dae|stl|STL))\"", urdf):
            mesh_description = m[0].split("/")[0]
            mesh_repo: Optional[Path] = None
            mesh_desc = ""
            for k_desc in repo_paths:
                if mesh_description in k_desc:
                    mesh_repo: Path = repo_paths[k_desc]
                    mesh_desc = k_desc
                    break
            if mesh_repo is None:
                raise Exception(f"Couldn't find local repo for: {m[0]}")
            if "/" in mesh_desc:
                mesh_src = mesh_repo.joinpath(mesh_desc.replace(mesh_desc.split("/")[-1], "")).joinpath(m[0])
            else:
                mesh_src = mesh_repo.joinpath(m[0])
            assert mesh_src.exists(), f"Not found: {mesh_src}"
            mesh_dst = dst_root.joinpath(m[0])
            if not mesh_dst.parent.exists():
                mesh_dst.parent.mkdir(parents=True)
            # Copy the mesh file to the Unity project.
            copy_file(src=str(mesh_src.resolve()), dst=str(mesh_dst.resolve()))
        if not self._quiet:
            print("Copied the .urdf and the meshes to the Unity project.")
        return urdf_dst

    def xacro_to_urdf(self, xacro_path: Path, repo_paths: Dict[str, Path], args: Dict[str, str] = None) -> Path:
        """
        Convert a local .xacro file to a .urdf file.

        :param xacro_path: The path to the local .xacro file.
        :param args: Names and values for the `arg` tags in the .xacro file.
        :param repo_paths: Local paths to all required repos. Key = The description infix. Value = The local repo path.

        :return: The path to the .urdf file.
        """

        if not RobotCreator.TEMP_ROOT.exists():
            RobotCreator.TEMP_ROOT.mkdir(parents=True)

        xacro = xacro_path.read_text(encoding="utf-8")

        # Set the args.
        if args is None:
            args = {"gazebo": 'false'}
        for k in args:
            xacro = re.sub('<xacro:arg name="' + k + '" default="(.*)"',
                           f'<xacro:arg name="{k}" default="{args[k]}"', xacro)
        xacro = re.sub(r'<xacro:include (.*)gazebo(.*?)/>', "", xacro)

        # Put all required .xacro files in a temporary directory.
        xacro_dir = RobotCreator.TEMP_ROOT.joinpath("xacro")
        if not xacro_dir.exists():
            xacro_dir.mkdir(parents=True)
        x = xacro_dir.joinpath(xacro_path.name)
        x.write_text(xacro, encoding="utf-8")

        xacros: List[Path] = [x]
        checked: List[Path] = []
        while len(xacros) > 0:
            xp = xacros.pop(0)
            checked.append(xp)
            xacro = xp.read_text(encoding="utf-8")
            for description in re.findall(r"\$\(find (.*?)\)", xacro, flags=re.MULTILINE):
                xacro_repo: Optional[Path] = None
                desc = ""
                for k_desc in repo_paths:
                    if description in k_desc:
                        xacro_repo = repo_paths[k_desc]
                        desc = k_desc
                        break
                assert xacro_repo is not None, f"Couldn't find: {description} in {xacro_repo} for {xp}"
                src_urdf_dir = xacro_repo.joinpath(desc).joinpath("urdf")
                for root_dir, dirs, files in walk(str(src_urdf_dir.resolve())):
                    for f in files:
                        src = Path(root_dir).joinpath(f)
                        if src.is_file() and src.suffix == ".xacro":
                            dst = xacro_dir.joinpath(src.name)
                            if not dst.exists():
                                copy_file(src=str(src.resolve()), dst=str(dst.resolve()))
                            if src not in xacros and src not in checked:
                                xacros.append(src)
        if not self._quiet:
            print("Copied all required xacro files to a temp directory.")
        # "Repair" all of the required .xacro files.
        for f in xacro_dir.iterdir():
            if f.is_file() and f.suffix == ".xacro":
                xacro = f.read_text(encoding="utf-8")
                xacro = re.sub(r"include filename=\"\$\((.*)\)/(.*)/(.*)\"", r'include filename="\3"', xacro)
                f.write_text(xacro, encoding="utf-8")
        # Finally, create the .urdf file.
        cwd = getcwd()
        chdir(str(RobotCreator.TEMP_ROOT.joinpath("xacro").resolve()))
        urdf_name = x.name.replace(".urdf.xacro", ".urdf").replace(".xacro", ".urdf")
        xacro_call = ["source", "/opt/ros/melodic/setup.bash", "&&",
                      "rosrun", "xacro", "xacro", "-o", urdf_name, x.name]
        if system() == "Windows":
            xacro_call.insert(0, "wsl")
        call(xacro_call)
        urdf_path = Path(f"../../Assets/robots/{urdf_name}")
        if urdf_path.exists():
            urdf_path.unlink()
        move_file(src=str(x.parent.joinpath(urdf_name).resolve()), dst=str(urdf_path.resolve()))
        if not self._quiet:
            print(f"Created {str(urdf_path.resolve())}")
        urdf_path = Path(str(urdf_path.resolve()))
        chdir(cwd)
        # Delete temp xacro files.
        remove_tree(str(xacro_dir.resolve()))
        assert urdf_path.exists(), f"Not found: {urdf_path.resolve()}"
        return urdf_path.resolve()

    def urdf_to_prefab(self, urdf_path: Path, immovable: bool = True, up: str = "y") -> Path:
        """
        Convert a .urdf file to Unity prefab.

        The .urdf file must already exist on this machine and its meshes must be at the expected locations.

        :param urdf_path: The path to the .urdf file.
        :param immovable: If True, the base of the robot will be immovable by default (see the `set_immovable` command).
        :param up: The up direction. Used for importing the .urdf into Unity. Options: "y" or "z".

        :return: The path to the .prefab file.
        """

        # Get the expected name of the robot.
        urdf = urdf_path.read_text(encoding="utf-8")
        name = re.search(r'<robot name="(.*?)"', urdf, flags=re.MULTILINE).group(1).strip()

        urdf_call = self.get_base_unity_call()[:]
        urdf_call.extend(["-executeMethod", "Creator.CreatePrefab",
                          f"-urdf='{str(urdf_path.resolve())}'",
                          f"-immovable={'true' if immovable else 'false'}",
                          f"-up={up}"])
        if not self._quiet:
            print("Creating a .prefab from a .urdf file...")
        call(urdf_call)
        RobotCreator._check_log()
        prefab_path = self._project_path.joinpath(f"Assets/prefabs/{name}.prefab")
        assert prefab_path.exists(), f"Prefab not found: {prefab_path}"
        if not self._quiet:
            print("...Done!")
        return prefab_path

    def prefab_to_asset_bundles(self, name: str) -> Dict[str, Path]:
        """
        Create asset bundles from a prefab.

        :param name: The name of the robot (minus the .prefab extension).

        :return: A dictionary. Key = The system platform. Value = The path to the asset bundle as a Path object.
        """

        asset_bundles_call = self.get_base_unity_call()[:]
        asset_bundles_call.extend(["-executeMethod", "Creator.CreateAssetBundles",
                                   f"-robot='{name}'"])
        if not self._quiet:
            print("Creating asset bundles...")
        call(asset_bundles_call)
        RobotCreator._check_log()
        # Verify that the asset bundles exist.
        asset_bundles_root_dir = self._project_path.joinpath(f"Assets/asset_bundles/{name}")
        asset_bundle_paths: Dict[str, Path] = dict()
        for build_target in UNITY_TO_SYSTEM:
            asset_bundle_path = asset_bundles_root_dir.joinpath(f"{build_target}/{name}")
            asset_bundle_path = Path(str(Path(asset_bundle_path.resolve())))
            assert asset_bundle_path.exists(), f"Couldn't find asset bundle: {asset_bundle_path.resolve()}"
            asset_bundle_paths[UNITY_TO_SYSTEM[build_target]] = asset_bundle_path
        if not self._quiet:
            print("...Done!")
        return asset_bundle_paths

    def get_unity_project(self) -> Path:
        """
        Build the asset_bundle_creator Unity project.

        :return The path to the asset_bundle_creator Unity project.
        """

        unity_project_path = self.get_project_path()

        # If the project already exists, stop.
        if unity_project_path.exists():
            assert unity_project_path.joinpath("Assets").exists(), f"There is a directory at {unity_project_path} but it's not a Unity project."
            return unity_project_path

        # Clone the repo.
        cwd = getcwd()
        chdir(str(Path.home().resolve()))
        call(["git", "clone", "https://github.com/alters-mit/robot_creator"])
        chdir(cwd)
        assert unity_project_path.exists(), f"Can't find project path: {unity_project_path}"
        return unity_project_path

    @staticmethod
    def get_project_path() -> Path:
        """
        :return: The expected path of the Unity project.
        """

        return Path.home().joinpath("robot_creator")

    @staticmethod
    def _check_log() -> None:
        """
        Check the Editor log for errors.
        """

        log = EDITOR_LOG_PATH.read_text(encoding="utf-8")
        if "failure" in log.lower() or "exception" in log.lower():
            raise Exception(f"There are errors in the Editor log!")

    @staticmethod
    def _page_to_raw(url: str) -> str:
        """
        Convert the URL of a GitHub page to the URL of the corresponding text file.

        :param url: A URL to a GitHub page.

        :return: The URL to the corresponding text file.
        """

        if "https://github.com" in url:
            return re.sub(r"https://github\.com/(.*)/blob/(.*)", r"https://raw.githubusercontent.com/\1/\2", url)
        elif "https://raw.githubusercontent.com" in url:
            return url
        else:
            raise Exception(f"Unexpected URL: {url}")

    @staticmethod
    def _raw_to_page(url: str):
        """
        Convert the URL of a raw text file to the corresponding GitHub page.

        :param url: A URL to a text file page.

        :return: The URL to the corresponding GitHub page.
        """

        if "https://github.com" in url:
            return url
        elif "https://raw.githubusercontent.com" in url:
            return re.sub(r"https://raw\.githubusercontent\.com/(.*)/master/(.*)",
                          r"https://github.com/\1/blob/master/\2", url)
        else:
            raise Exception(f"Unexpected URL: {url}")

    @staticmethod
    def _get_repo_url(url: str) -> str:
        """
        :param url: The URL of the .urdf or .xacro file.

        :return: The base repo of a .urdf or .xacro file.
        """

        page_url = RobotCreator._raw_to_page(url=url)
        return re.sub(r"https://github\.com/(.*)/blob/(.*)", r"https://github.com/\1", page_url)

    @staticmethod
    def _get_repo_name(repo_url: str) -> str:
        """
        :param repo_url: The base URL of a repo.

        :return: The expected name of the repo.
        """

        return repo_url.split("/")[-1]

    @staticmethod
    def _get_local_repo_path(url: str) -> Path:
        """
        :param url: The URL of the .urdf or .xacro file.

        :return: The path to the local repo.
        """

        if not RobotCreator.TEMP_ROOT.exists():
            RobotCreator.TEMP_ROOT.mkdir(parents=True)

        repo_url = RobotCreator._get_repo_url(url=url)
        repo_name = RobotCreator._get_repo_name(repo_url=repo_url)
        return RobotCreator.TEMP_ROOT.joinpath(repo_name)

    @staticmethod
    def _get_description_infix(url: str) -> str:
        """
        :param url: The URL of the .urdf or .xacro file.

        :return: The string between the repo URL and the /urdf/ directory.
        """

        page = RobotCreator._raw_to_page(url=url)
        s = re.search(r"(.*)/blob/master/(.*)/urdf", page)
        if s is None:
            return re.search(r"(.*)/blob/master/((.*)_description)/", page).group(2)
        else:
            return s.group(2)



import numpy as np


class Bound:
    """
    Bounds data for a single object.
    """

    def __init__(self, front: np.array, back: np.array, left: np.array, right: np.array, top: np.array,
                 bottom: np.array, center: np.array):
        """
        :param front: The position of the front point.
        :param back: The position of the back point.
        :param left: The position of the left point.
        :param right: The position of the right point.
        :param top: The position of the top point.
        :param bottom: The position of the bottom point.
        :param center: The position of the center point.
        """

        """:field
        The position of the front point.
        """
        self.front: np.array = front
        """:field
        The position of the back point.
        """
        self.back: np.array = back
        """:field
        The position of the left point.
        """
        self.left: np.array = left
        """:field
        The position of the right point.
        """
        self.right: np.array = right
        """:field
        The position of the top point.
        """
        self.top: np.array = top
        """:field
        The position of the bottom point.
        """
        self.bottom: np.array = bottom
        """:field
        The position of the center point.
        """
        self.center: np.array = center



from typing import List
from tdw.add_ons.model_verifier.model_tests.model_test import ModelTest
from tdw.output_data import OutputData, LogMessage


class ModelReport(ModelTest):
    """
    Send `send_model_report` and get a basic report on the model.
    """

    def start(self) -> List[dict]:
        """
        :return: A list of commands to start the test.
        """

        return [{"$type": "send_model_report",
                 "name": self._record.name,
                 "url": self._record.get_url(),
                 "id": 0,
                 "scale_factor": self._record.scale_factor}]

    def on_send(self, resp: List[bytes]) -> List[dict]:
        """
        :param resp: The response from the build.

        :return: A list of commands to continue or end the test.
        """

        self.done = True
        for i in range(len(resp) - 1):
            r_id = OutputData.get_data_type_id(resp[i])
            if r_id == "logm":
                log_message = LogMessage(resp[i])
                self.reports.append(log_message.get_message_type() + ": " + log_message.get_message())
        return []




from typing import List, Dict, Optional, Tuple
import numpy as np
from tdw.output_data import OutputData, Robot, Collision, EnvironmentCollision
from tdw.object_data.transform import Transform
from tdw.robot_data.joint_dynamic import JointDynamic
from tdw.collision_data.collision_obj_obj import CollisionObjObj
from tdw.collision_data.collision_obj_env import CollisionObjEnv


class RobotDynamic:
    """
    Dynamic data for a robot that can change per frame (such as the position of the robot, the angle of a joint, etc.)
    """

    """:class_var
    If the joint moved by less than this angle or distance since the previous frame, it's considered to be non-moving.
    """
    NON_MOVING: float = 0.001

    def __init__(self, robot_id: int, resp: List[bytes], body_parts: List[int], previous=None):
        """
        :param resp: The response from the build, which we assume contains `Robot` output data.
        :param robot_id: The ID of this robot.
        :param body_parts: The IDs of all body parts belonging to this robot.
        :param previous: If not None, the previous RobotDynamic data. Use this to determine if the joints are moving.
        """

        """:field
        The Transform data for this robot.
        """
        self.transform: Optional[Transform] = None
        """:field
        A dictionary of [dynamic joint data](joint_dynamic.md). Key = The ID of the joint.
        """
        self.joints: Dict[int, JointDynamic] = dict()
        """:field
        If True, this robot is immovable.
        """
        self.immovable: bool = False
        """:field
        A dictionary of collisions between one of this robot's [body parts (joints or non-moving)](robot_static.md) and another object.
        Key = A tuple where the first element is the body part ID and the second element is the object ID.
        Value = A list of [collision data.](../collision_data/collision_obj_obj.md)
        """
        self.collisions_with_objects: Dict[Tuple[int, int], List[CollisionObjObj]] = dict()
        """:field
        A dictionary of collisions between two of this robot's [body parts](robot_static.md).
        Key = An unordered tuple of two body part IDs.
        Value = A list of [collision data.](../collision_data/collision_obj_obj.md)
        """
        self.collisions_with_self: Dict[Tuple[int, int], List[CollisionObjObj]] = dict()
        """:field
        A dictionary of collisions between one of this robot's [body parts](robot_static.md) and the environment (floors, walls, etc.).
        Key = The ID of the body part.
        Value = A list of [environment collision data.](../collision_data/collision_obj_env.md)
        """
        self.collisions_with_environment: Dict[int, List[CollisionObjEnv]] = dict()
        for i in range(len(resp) - 1):
            r_id = OutputData.get_data_type_id(resp[i])
            if r_id == "robo":
                robot: Robot = Robot(resp[i])
                if robot.get_id() == robot_id:
                    self.transform = Transform(position=np.array(robot.get_position()),
                                               rotation=np.array(robot.get_rotation()),
                                               forward=np.array(robot.get_forward()))
                    self.immovable = robot.get_immovable()
                    # Get dynamic data for each joint.
                    for j in range(robot.get_num_joints()):
                        joint = JointDynamic(robot=robot, joint_index=j)
                        # Determine if the joint is currently moving.
                        if previous is not None:
                            previous: RobotDynamic
                            previous_joint: JointDynamic = previous.joints[joint.joint_id]
                            for k in range(len(previous_joint.angles)):
                                if np.linalg.norm(previous_joint.angles[k] - joint.angles[k]) > RobotDynamic.NON_MOVING:
                                    joint.moving = True
                                    break
                        self.joints[joint.joint_id] = joint
            # Record collisions between myself and my joints or with another object.
            elif r_id == "coll":
                collision = Collision(resp[i])
                collider_id: int = collision.get_collider_id()
                collidee_id: int = collision.get_collidee_id()
                # Record collisions between one of my body parts and another of my body parts.
                if collider_id in body_parts and collidee_id in body_parts:
                    key: Tuple[int, int] = (collider_id, collidee_id)
                    c = CollisionObjObj(collision)
                    # Record this collision.
                    if key not in self.collisions_with_self:
                        self.collisions_with_self[key] = [c]
                    else:
                        self.collisions_with_self[key].append(c)
                # Record collisions between one of my body parts and another object.
                elif collider_id in body_parts or collidee_id in body_parts:
                    # The body part is the first element in the tuple.
                    if collider_id in body_parts:
                        key: Tuple[int, int] = (collider_id, collidee_id)
                    else:
                        key: Tuple[int, int] = (collidee_id, collider_id)
                    # Record this collision.
                    c = CollisionObjObj(collision)
                    if key not in self.collisions_with_self:
                        self.collisions_with_objects[key] = [c]
                    else:
                        self.collisions_with_objects[key].append(c)
            elif r_id == "enco":
                collision = EnvironmentCollision(resp[i])
                object_id = collision.get_object_id()
                if object_id in body_parts:
                    c = CollisionObjEnv(collision)
                    if object_id not in self.collisions_with_environment:
                        self.collisions_with_environment[object_id] = [c]
                    else:
                        self.collisions_with_environment[object_id].append(c)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FBOutput

import tdw.flatbuffers

class StaticRigidbodyData(object):
    __slots__ = ['_tab']

    # StaticRigidbodyData
    def Init(self, buf, pos):
        self._tab = tdw.flatbuffers.table.Table(buf, pos)

    # StaticRigidbodyData
    def Id(self): return self._tab.Get(tdw.flatbuffers.number_types.Int32Flags, self._tab.Pos + tdw.flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # StaticRigidbodyData
    def Mass(self): return self._tab.Get(tdw.flatbuffers.number_types.Float32Flags, self._tab.Pos + tdw.flatbuffers.number_types.UOffsetTFlags.py_type(4))
    # StaticRigidbodyData
    def Kinematic(self): return self._tab.Get(tdw.flatbuffers.number_types.BoolFlags, self._tab.Pos + tdw.flatbuffers.number_types.UOffsetTFlags.py_type(8))
    # StaticRigidbodyData
    def DynamicFriction(self): return self._tab.Get(tdw.flatbuffers.number_types.Float32Flags, self._tab.Pos + tdw.flatbuffers.number_types.UOffsetTFlags.py_type(12))
    # StaticRigidbodyData
    def StaticFriction(self): return self._tab.Get(tdw.flatbuffers.number_types.Float32Flags, self._tab.Pos + tdw.flatbuffers.number_types.UOffsetTFlags.py_type(16))
    # StaticRigidbodyData
    def Bounciness(self): return self._tab.Get(tdw.flatbuffers.number_types.Float32Flags, self._tab.Pos + tdw.flatbuffers.number_types.UOffsetTFlags.py_type(20))

def CreateStaticRigidbodyData(builder, id, mass, kinematic, dynamicFriction, staticFriction, bounciness):
    builder.Prep(4, 24)
    builder.PrependFloat32(bounciness)
    builder.PrependFloat32(staticFriction)
    builder.PrependFloat32(dynamicFriction)
    builder.Pad(3)
    builder.PrependBool(kinematic)
    builder.PrependFloat32(mass)
    builder.PrependInt32(id)
    return builder.Offset()



from typing import Optional, Dict
import numpy as np
from tdw.collision_data.collision_base import CollisionBase
from tdw.collision_data.collision_obj_obj import CollisionObjObj
from tdw.collision_data.collision_obj_env import CollisionObjEnv
from tdw.object_data.rigidbody import Rigidbody
from tdw.physics_audio.collision_audio_type import CollisionAudioType
from tdw.physics_audio.object_audio_static import ObjectAudioStatic


class CollisionAudioEvent:
    """
    Data for a collision audio event.
    Includes collision data as well as the "primary" and "secondary" objects and the type of audio event.
    """

    """:class_var
    If the angular velocity is this or greater, the event is a roll, not a scrape.
    """
    ROLL_ANGULAR_VELOCITY: float = 0.5
    """:class_var
    If the area of the collision increases by at least this factor during a stay event, the collision is actually an impact.
    """
    IMPACT_AREA_RATIO: float = 5
    """:class_var
    On a stay event, if the previous area is None and the current area is greater than this, the collision is actually an impact. 
    """
    IMPACT_AREA_NEW_COLLISION: float = 1e-5

    def __init__(self, collision: CollisionBase, object_0_static: ObjectAudioStatic, object_0_dynamic: Rigidbody,
                 previous_areas: Dict[int, float], object_1_static: ObjectAudioStatic = None,
                 object_1_dynamic: Rigidbody = None):
        """
        :param collision: [The collision data.](../collision_data/collision_base.md)
        :param object_0_static: [Static data](object_audio_static.md) for the first object.
        :param object_0_dynamic: [Dynamic data](../object_data/rigidbody.md) for the first object.
        :param previous_areas: Areas of collisions from the previous frame.
        :param object_1_static: [Static data](object_audio_static.md) for the second object. If this is an environment collision, the value of this parameter should be `None`.
        :param object_1_dynamic: [Dynamic data](../object_data/rigidbody.md) for the second object. If this is an environment collision, the value of this parameter should be `None`.
        """

        """:field
        [The collision data.](../collision_data/collision_base.md)
        """
        self.collision: CollisionBase = collision
        """:field
        The area of the collision contact points.
        """
        self.area: float = 0
        """:field
        [The collision audio event type.](collision_audio_type.md)
        """
        self.collision_type: CollisionAudioType = CollisionAudioType.none
        """:field
        The ID of the primary object.
        """
        self.primary_id: int = object_0_static.object_id
        """:field
        The ID of the secondary object. If this is an environment collision, the value of this field is `None`.
        """
        self.secondary_id: Optional[int] = None
        """:field
        A value to mark the overall "significance" of the collision event.
        """
        self.magnitude: float = 0
        """:field
        The velocity vector.
        """
        self.velocity: np.array = np.array([0, 0, 0])
        if collision.state == "exit":
            return

        if isinstance(collision, CollisionObjObj):
            self.velocity = collision.relative_velocity
            self.magnitude = np.linalg.norm(self.velocity)
            valid_event = self.magnitude > 0.01
            obj_obj: bool = True
            if object_1_static is None:
                raise Exception("object_1_static is None but this is an object-object collision.")
        elif isinstance(collision, CollisionObjEnv):
            self.velocity = object_0_dynamic.velocity
            self.magnitude = np.linalg.norm(self.velocity)
            valid_event = self.magnitude > 0.01
            obj_obj = False
        else:
            raise Exception(f"Invalid collision type: {collision}")
        if not valid_event:
            return
        # Initially set this as an impact so we can find the previous area.
        self._set_as_impact(obj_obj=obj_obj, object_0_static=object_0_static, object_1_static=object_1_static)
        # Set the area.
        self.area = self._get_contact_area()
        # Get the previous area, if any.
        if self.primary_id in previous_areas:
            previous_area = previous_areas[self.primary_id]
        else:
            previous_area = None
        if self.collision.state == "stay":
            if previous_area is None:
                if self.area > CollisionAudioEvent.IMPACT_AREA_NEW_COLLISION:
                    self._set_as_impact(obj_obj=obj_obj,
                                        object_0_static=object_0_static,
                                        object_1_static=object_1_static)
                else:
                    self.collision_type = CollisionAudioType.none
            # This is a scrape or a roll.
            elif previous_area > 0 and self.area / previous_area < CollisionAudioEvent.IMPACT_AREA_RATIO:
                self._set_as_impact(obj_obj=obj_obj, object_0_static=object_0_static, object_1_static=object_1_static)
                # Get the angular velocity of the primary object.
                if obj_obj:
                    if object_1_dynamic is None:
                        raise Exception("object_1_dynamic is None but this is an object-object collision.")
                    # Set the primary and secondary bodies based on speed. Assume that the slower object is the surface.
                    if np.linalg.norm(object_0_dynamic.velocity) > np.linalg.norm(object_1_dynamic.velocity):
                        self.primary_id = object_0_static.object_id
                        self.secondary_id = object_1_static.object_id
                        angular_velocity = object_0_dynamic.angular_velocity
                    else:
                        self.primary_id = object_1_static.object_id
                        self.secondary_id = object_0_static.object_id
                        angular_velocity = object_1_dynamic.angular_velocity
                else:
                    angular_velocity = object_0_dynamic.angular_velocity
                # If the primary object has a high angular velocity, this is a roll.
                if np.linalg.norm(angular_velocity) > CollisionAudioEvent.ROLL_ANGULAR_VELOCITY:
                    # TODO set this to CollisionAudioType.roll once we have roll sounds.
                    self.collision_type = CollisionAudioType.impact
                # If the primary object has a low angular velocity, this is a scrape.
                else:
                    self.collision_type = CollisionAudioType.scrape
            else:
                self.collision_type = CollisionAudioType.none

    def _get_contact_area(self) -> float:
        """
        :return: The area of all of the contact points.
        """

        # Source: https://stackoverflow.com/a/68115011
        points = np.array(self.collision.points)
        edges = points[1:] - points[0:1]
        return sum(np.linalg.norm(np.cross(edges[:-1], edges[1:], axis=1), axis=1) / 2)

    def _set_as_impact(self, obj_obj: bool, object_0_static: ObjectAudioStatic, object_1_static: ObjectAudioStatic) -> None:
        """
        Set the primary (target) object and secondary (other) object for an impact event based on relative mass.

        :param obj_obj: If True, this is an object-object collision.
        :param object_0_static: Static data for the first object.
        :param object_1_static: Static data for the second object.
        """

        self.collision_type = CollisionAudioType.impact
        if not obj_obj:
            return
        elif object_0_static.mass < object_1_static.mass:
            self.primary_id = object_0_static.object_id
            self.secondary_id = object_1_static.object_id
        else:
            self.primary_id = object_1_static.object_id
            self.secondary_id = object_0_static.object_id



from typing import Dict, List, Union
import numpy as np
from tdw.output_data import OutputData, AvatarSimpleBody, ImageSensors
from tdw.tdw_utils import TDWUtils
from tdw.add_ons.third_person_camera_base import ThirdPersonCameraBase
from tdw.add_ons.avatar_body import AvatarBody
from tdw.object_data.transform import Transform
from tdw.object_data.rigidbody import Rigidbody


class EmbodiedAvatar(ThirdPersonCameraBase):
    """
    An `EmbodiedAvatar` is an avatar with a physical body. The body has a simple shape and responds to physics (just like objects and robots).
    """

    def __init__(self, avatar_id: str = None, position: Dict[str, float] = None, rotation: Dict[str, float] = None,
                 field_of_view: int = None, color: Dict[str, float] = None, body: AvatarBody = AvatarBody.capsule,
                 scale_factor: Dict[str, float] = None, mass: float = 80, dynamic_friction: float = 0.3,
                 static_friction: float = 0.3, bounciness: float = 0.7, drag: float = 1, angular_drag: float = 0.5):
        """
        :param avatar_id: The ID of the avatar. If None, a random ID is generated.
        :param position: The initial position of the avatar. If None, defaults to `{"x": 0, "y": 0, "z": 0}`.
        :param rotation: The initial rotation of the avatar. Can be Euler angles (keys are `(x, y, z)`) or a quaternion (keys are `(x, y, z, w)`). If None, defaults to `{"x": 0, "y": 0, "z": 0}`.
        :param field_of_view: The initial field of view.
        :param color: The color of the avatar as an `r, g, b, a` dictionary where each value is between 0 and 1. Can be None.
        :param body: [The body of the avatar.](avatar_body.md)
        :param scale_factor: Scale the avatar by this factor. Can be None.
        :param mass: The mass of the avatar.
        :param dynamic_friction: The dynamic friction coefficient of the avatar.
        :param static_friction: The static friction coefficient of the avatar.
        :param bounciness: The bounciness of the avatar.
        """

        super().__init__(avatar_id=avatar_id, position=position, rotation=rotation, field_of_view=field_of_view)
        """:field
        [Transform data](../object_data/transform.md) for the avatar.
        """
        self.transform: Transform = Transform(position=np.array([0, 0, 0]),
                                              rotation=np.array([0, 0, 0, 0]),
                                              forward=np.array([0, 0, 0]))
        """:field
        [Rigidbody data](../object_data/rigidbody.md) for the avatar.
        """
        self.rigidbody: Rigidbody = Rigidbody(velocity=np.array([0, 0, 0]),
                                              angular_velocity=np.array([0, 0, 0]),
                                              sleeping=True)
        """:field
        The rotation of the camera as an [x, y, z, w] numpy array.
        """
        self.camera_rotation: np.array = np.array([0, 0, 0, 0])
        """:field
        If True, the avatar is currently moving or turning.
        """
        self.is_moving: bool = False
        self._body: AvatarBody = body
        self._mass: float = mass
        self._dynamic_friction: float = dynamic_friction
        self._static_friction: float = static_friction
        self._bounciness: float = bounciness
        self._drag: float = drag
        self._angular_drag: float = angular_drag
        self._color: Dict[str, float] = color
        self._scale_factor: Dict[str, float] = scale_factor

    def get_initialization_commands(self) -> List[dict]:
        commands = super().get_initialization_commands()
        commands.extend([{"$type": "change_avatar_body",
                          "body_type": self._body.name.title(),
                          "avatar_id": self.avatar_id},
                         {"$type": "set_avatar_mass",
                          "mass": self._mass,
                          "avatar_id": self.avatar_id},
                         {"$type": "set_avatar_physic_material",
                          "dynamic_friction": self._dynamic_friction,
                          "static_friction": self._static_friction,
                          "bounciness": self._bounciness,
                          "avatar_id": self.avatar_id},
                         {"$type": "set_avatar_drag",
                          "angular_drag": 0.5,
                          "avatar_id": self.avatar_id,
                          "drag": 1},
                         {"$type": "send_avatars",
                          "frequency": "always"},
                         {"$type": "send_image_sensors",
                          "frequency": "always"},
                         {"$type": "set_avatar_drag",
                          "drag": self._drag,
                          "angular_drag": self._angular_drag,
                          "avatar_id": self.avatar_id}])
        if self._color is not None:
            if "a" in self._color and self._color["a"] < 1:
                commands.append({"$type": "enable_avatar_transparency",
                                 "avatar_id": self.avatar_id})
            commands.append({"$type": "set_avatar_color",
                             "color": self._color,
                             "avatar_id": self.avatar_id})
        if self._scale_factor is not None:
            commands.append({"$type": "scale_avatar",
                             "scale_factor": self._scale_factor,
                             "avatar_id": self.avatar_id})
        return commands

    def apply_force(self, force: Union[float, int, Dict[str, float], np.ndarray]) -> None:
        """
        Apply a force to the avatar to begin moving it.

        :param force: The force. If float: apply a force along the forward (positive value) or backward (negative value) directional vector of the avatar. If dictionary or numpy array: Apply a force defined by the x, y, z vector.
        """

        if isinstance(force, float) or isinstance(force, int):
            self.commands.append({"$type": "move_avatar_forward_by",
                                  "magnitude": force,
                                  "avatar_id": self.avatar_id})
        elif isinstance(force, dict):
            force_arr = TDWUtils.vector3_to_array(force)
            force_magnitude = float(np.linalg.norm(force_arr))
            self.commands.append({"$type": "apply_force_to_avatar",
                                  "magnitude": force_magnitude,
                                  "direction": force_arr / force_magnitude,
                                  "avatar_id": self.avatar_id})
        elif isinstance(force, np.ndarray):
            force_magnitude = float(np.linalg.norm(force))
            self.commands.append({"$type": "apply_force_to_avatar",
                                  "magnitude": force_magnitude,
                                  "direction": force / force_magnitude,
                                  "avatar_id": self.avatar_id})
        else:
            raise Exception(f"Invalid type: {force}")
        self.is_moving = True

    def apply_torque(self, torque: float) -> None:
        """
        Apply a torque to the avatar so that it starts turning.

        :param torque: The torque. Positive value = clockwise rotation.
        """

        self.commands.append({"$type": "turn_avatar_by",
                              "torque": torque,
                              "avatar_id": self.avatar_id})
        self.is_moving = True

    def set_drag(self, drag: float = 1, angular_drag: float = 0.5) -> None:
        """
        Set the drag of the avatar. Increase this to force the avatar to slow down.

        :param drag: The drag of the rigidbody. A higher drag value will cause the avatar slow down faster.
        :param angular_drag: The angular drag of the rigidbody. A higher angular drag will cause the avatar's rotation to slow down faster.
        """

        self.commands.append({"$type": "set_avatar_drag",
                              "drag": drag,
                              "angular_drag": angular_drag,
                              "avatar_id": self.avatar_id})

    def rotate_camera(self, rotation: Dict[str, float]) -> None:
        """
        Rotate the camera.

        :param rotation: Rotate the camera by these angles (in degrees). Keys are `"x"`, `"y"`, `"z"` and correspond to `(pitch, yaw, roll)`.
        """

        for q, axis in zip(["x", "y", "z"], ["pitch", "yaw", "roll"]):
            self.commands.append({"$type": "rotate_sensor_container_by",
                                  "axis": axis,
                                  "angle": rotation[q],
                                  "avatar_id": self.avatar_id})

    def look_at(self, target: Union[int, Dict[str, float], np.ndarray]) -> None:
        """
        Look at a target object or position.

        :param target: The target. If int: an object ID. If a dictionary or numpy array: an x, y, z position.
        """

        if isinstance(target, int):
            self.commands.append({"$type": "look_at",
                                  "object_id": target,
                                  "use_centroid": True,
                                  "avatar_id": self.avatar_id})
        elif isinstance(target, dict):
            self.commands.append({"$type": "look_at_position",
                                  "position": target,
                                  "avatar_id": self.avatar_id})
        elif isinstance(target, np.ndarray):
            self.commands.append({"$type": "look_at_position",
                                  "position": TDWUtils.array_to_vector3(target),
                                  "avatar_id": self.avatar_id})
        else:
            raise Exception(f"Invalid type: {target}")

    def reset_camera(self) -> None:
        """
        Reset the rotation of the camera.
        """

        self.commands.append({"$type": "reset_sensor_container_rotation",
                              "avatar_id": self.avatar_id})

    def on_send(self, resp: List[bytes]) -> None:
        for i in range(len(resp) - 1):
            r_id = OutputData.get_data_type_id(resp[i])
            # Update my state.
            if r_id == "avsb":
                avsb = AvatarSimpleBody(resp[i])
                if avsb.get_avatar_id() == self.avatar_id:
                    # Update the transform data.
                    self.transform.position = np.array(avsb.get_position())
                    self.transform.rotation = np.array(avsb.get_rotation())
                    self.transform.forward = np.array(avsb.get_forward())
                    # Update the rigidbody data.
                    self.rigidbody.velocity = np.array(avsb.get_velocity())
                    self.rigidbody.angular_velocity = np.array(avsb.get_angular_velocity())
                    self.rigidbody.sleeping = np.array(avsb.get_sleeping())
                    # Update whether the avatar is moving.
                    self.is_moving = not self.rigidbody.sleeping
            # Update the rotation of the camera.
            elif r_id == "imse":
                imse = ImageSensors(resp[i])
                if imse.get_avatar_id() == self.avatar_id:
                    self.camera_rotation = np.array(imse.get_sensor_rotation(0))

    def _get_avatar_type(self) -> str:
        return "A_Simple_Body"



from enum import Enum


class ScrapeMaterial(Enum):
    """
    An audio material used for a scrape event.
    """

    plywood = 0
    ceramic = 1
    pvc = 2
    rough_wood = 3
    acrylic = 4
    sanded_acrylic = 5
    vinyl = 6
    poplar_wood = 7
    bass_wood = 8
    polycarbonate = 9
    polyethylene = 10
    sandpaper = 11



from typing import Optional, Union, Tuple
from pathlib import Path
import os
from subprocess import check_output, Popen, call
import re
from psutil import pid_exists


class AudioUtils:
    """
    Utility class for recording audio in TDW using [fmedia](https://stsaz.github.io/fmedia/).

    Usage:

    ```python
    from tdw.audio_utils import AudioUtils
    from tdw.controller import Controller

    c = Controller()

    initialize_trial()  # Your code here.

    # Begin recording audio. Automatically stop recording at 10 seconds.
    AudioUtils.start(output_path="path/to/file.wav", until=(0, 10))

    do_trial()  # Your code here.

    # Stop recording.
    AudioUtils.stop()
    ```
    """

    # The process ID of the audio recorder.
    RECORDER_PID: Optional[int] = None
    # The audio capture device.
    DEVICE: Optional[str] = None

    @staticmethod
    def get_system_audio_device() -> str:
        """
        :return: The audio device that can be used to capture system audio.
        """

        devices = check_output(["fmedia", "--list-dev"]).decode("utf-8").split("Capture:")[1]
        dev_search = re.search("device #(.*): Stereo Mix", devices, flags=re.MULTILINE)
        assert dev_search is not None, "No suitable audio capture device found:\n" + devices
        return dev_search.group(1)

    @staticmethod
    def start(output_path: Union[str, Path], until: Optional[Tuple[int, int]] = None) -> None:
        """
        Start recording audio.

        :param output_path: The path to the output file.
        :param until: If not None, fmedia will record until `minutes:seconds`. The value must be a tuple of 2 integers. If None, fmedia will record until you send `AudioUtils.stop()`.
        """

        if isinstance(output_path, str):
            p = Path(output_path).resolve()
        else:
            p = output_path

        # Create the directory.
        if not p.parent.exists():
            p.parent.mkdir(parents=True)

        # Set the capture device.
        if AudioUtils.DEVICE is None:
            AudioUtils.DEVICE = AudioUtils.get_system_audio_device()
        fmedia_call = ["fmedia",
                       "--record",
                       f"--dev-capture={AudioUtils.DEVICE}",
                       f"--out={str(p.resolve())}",
                       "--globcmd=listen"]
        # Automatically stop recording.
        if until is not None:
            fmedia_call.append(f"--until={str(until[0]).zfill(2)}:{str(until[1]).zfill(2)}")
        with open(os.devnull, "w+") as f:
            AudioUtils.RECORDER_PID = Popen(fmedia_call,
                                            stderr=f).pid

    @staticmethod
    def stop() -> None:
        """
        Stop recording audio (if any fmedia process is running).
        """

        if AudioUtils.RECORDER_PID is not None:
            with open(os.devnull, "w+") as f:
                call(['fmedia', '--globcmd=quit'], stderr=f, stdout=f)
            AudioUtils.RECORDER_PID = None

    @staticmethod
    def is_recording() -> bool:
        """
        :return: True if the fmedia recording process still exists.
        """

        return AudioUtils.RECORDER_PID is not None and pid_exists(AudioUtils.RECORDER_PID)



from collections import Counter
from typing import List
from io import BytesIO
from PIL import Image
from tdw.add_ons.model_verifier.model_tests.rotate_object_test import RotateObjectTest
from tdw.output_data import Images


class MissingMaterials(RotateObjectTest):
    """
    Check if any materials are missing.
    """

    def start(self) -> List[dict]:
        """
        :return: A list of commands to start the test.
        """

        commands = super().start()
        commands.insert(0, {"$type": "set_pass_masks",
                            "pass_masks": ["_img"]})
        return commands

    def _read_images(self, images: Images) -> None:
        image = Image.open(BytesIO(images.get_image(0)))
        # Source: https://stackoverflow.com/a/59709420
        colors = Counter(image.getdata())
        unique_colors = set(colors)
        if MissingMaterials.PINK in unique_colors:
            self.reports.append("Missing materials")
            self.done = True

    def on_send(self, resp: List[bytes]) -> List[dict]:
        """
        :param resp: The response from the build.

        :return: A list of commands to continue or end the test.
        """

        commands = super().on_send(resp=resp)
        if self.done:
            commands.extend(RotateObjectTest._get_end_commands())
        return commands



import pkg_resources
import json
from typing import Dict


class FluidType:
    """
    Metadata for a single fluid type. Specifies values for Flex fluid containers.
    """

    def __init__(self, fluid_type: str, viscosity: float, adhesion: float, cohesion: float):
        """
        :param fluid_type: The name of the fluid type.
        :param viscosity: The viscosity value of the fluid.
        :param adhesion: The adhesion value of the fluid.
        :param cohesion: The cohesion value of the fluid.
        """

        """:field
        The name of the fluid type.
        """
        self.fluid_type: str = fluid_type
        """:field
        The viscosity value of the fluid.
        """
        self.viscosity: float = viscosity
        """:field
        The adhesion value of the fluid.
        """
        self.adhesion: float = adhesion
        """:field
        The cohesion value of the fluid.
        """
        self.cohesion: float = cohesion


def __load() -> Dict[str, FluidType]:
    """
    :return: The default fluid type data.
    """

    fluid_types: Dict[str, FluidType] = {}
    with open(pkg_resources.resource_filename(__name__, "fluid_types.json"), "rt") as f:
        fluid_type_data = json.load(f)
        for ft in fluid_type_data:
            fluid_types[ft] = FluidType(fluid_type=ft,
                                        viscosity=fluid_type_data[ft]["viscosity"],
                                        adhesion=fluid_type_data[ft]["adhesion"],
                                        cohesion=fluid_type_data[ft]["cohesion"])
    return fluid_types

# The default fluid types. Key = The name of the fluid.
FLUID_TYPES: Dict[str, FluidType] = __load()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FBOutput

import tdw.flatbuffers

class EmptyObjects(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsEmptyObjects(cls, buf, offset):
        n = tdw.flatbuffers.encode.Get(tdw.flatbuffers.packer.uoffset, buf, offset)
        x = EmptyObjects()
        x.Init(buf, n + offset)
        return x

    # EmptyObjects
    def Init(self, buf, pos):
        self._tab = tdw.flatbuffers.table.Table(buf, pos)

    # EmptyObjects
    def Ids(self, j):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(tdw.flatbuffers.number_types.Int32Flags, a + tdw.flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # EmptyObjects
    def IdsAsNumpy(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(tdw.flatbuffers.number_types.Int32Flags, o)
        return 0

    # EmptyObjects
    def IdsLength(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # EmptyObjects
    def Positions(self, j):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(tdw.flatbuffers.number_types.Float32Flags, a + tdw.flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # EmptyObjects
    def PositionsAsNumpy(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(tdw.flatbuffers.number_types.Float32Flags, o)
        return 0

    # EmptyObjects
    def PositionsLength(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

def EmptyObjectsStart(builder): builder.StartObject(2)
def EmptyObjectsAddIds(builder, ids): builder.PrependUOffsetTRelativeSlot(0, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(ids), 0)
def EmptyObjectsStartIdsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def EmptyObjectsAddPositions(builder, positions): builder.PrependUOffsetTRelativeSlot(1, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(positions), 0)
def EmptyObjectsStartPositionsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def EmptyObjectsEnd(builder): return builder.EndObject()



import numpy as np


class QuaternionUtils:
    """
    Helper functions for using quaternions.

    Quaternions are always numpy arrays in the following order: `[x, y, z, w]`.
    This is the order returned in all Output Data objects.

    Vectors are always numpy arrays in the following order: `[x, y, z]`.
    """

    """:class_var
    The global up directional vector.
    """
    UP = np.array([0, 1, 0])
    """:class_var
    The global forward directional vector.
    """
    FORWARD: np.array = np.array([0, 0, 1])
    """:class_var
    The quaternion identity rotation.
    """
    IDENTITY = np.array([0, 0, 0, 1])

    @staticmethod
    def get_inverse(q: np.array) -> np.array:
        """
        Source: https://referencesource.microsoft.com/#System.Numerics/System/Numerics/Quaternion.cs

        :param q: The quaternion.

        :return: The inverse of the quaternion.
        """

        x = q[0]
        y = q[1]
        z = q[2]
        w = q[3]

        ls = x * x + y * y + z * z + w * w
        inv = 1.0 / ls

        return np.array([-x * inv, -y * inv, -z * inv, w * inv])

    @staticmethod
    def multiply(q1: np.array, q2: np.array) -> np.array:
        """
        Multiply two quaternions.
        Source: https://stackoverflow.com/questions/4870393/rotating-coordinate-system-via-a-quaternion

        :param q1: The first quaternion.
        :param q2: The second quaternion.
        :return: The multiplied quaternion: `q1 * q2`
        """

        x1 = q1[0]
        y1 = q1[1]
        z1 = q1[2]
        w1 = q1[3]

        x2 = q2[0]
        y2 = q2[1]
        z2 = q2[2]
        w2 = q2[3]

        w = w1 * w2 - x1 * x2 - y1 * y2 - z1 * z2
        x = w1 * x2 + x1 * w2 + y1 * z2 - z1 * y2
        y = w1 * y2 + y1 * w2 + z1 * x2 - x1 * z2
        z = w1 * z2 + z1 * w2 + x1 * y2 - y1 * x2
        return np.array([x, y, z, w])

    @staticmethod
    def get_conjugate(q: np.array) -> np.array:
        """
        Source: https://stackoverflow.com/questions/4870393/rotating-coordinate-system-via-a-quaternion

        :param q: The quaternion.

        :return: The conjugate of the quaternion: `[-x, -y, -z, w]`
        """

        x = q[0]
        y = q[1]
        z = q[2]
        w = q[3]

        return np.array([-x, -y, -z, w])

    @staticmethod
    def multiply_by_vector(q: np.array, v: np.array) -> np.array:
        """
        Multiply a quaternion by a vector.
        Source: https://stackoverflow.com/questions/4870393/rotating-coordinate-system-via-a-quaternion

        :param q: The quaternion.
        :param v: The vector.

        :return: A directional vector calculated from: `q * v`
        """

        q2 = (v[0], v[1], v[2], 0.0)
        return QuaternionUtils.multiply(QuaternionUtils.multiply(q, q2), QuaternionUtils.get_conjugate(q))[:-1]

    @staticmethod
    def world_to_local_vector(position: np.array, origin: np.array, rotation: np.array) -> np.array:
        """
        Convert a vector position in absolute world coordinates to relative local coordinates.
        Source: https://answers.unity.com/questions/601062/what-inversetransformpoint-does-need-explanation-p.html

        :param position: The position vector in world coordinates.
        :param origin: The origin vector of the local space in world coordinates.
        :param rotation: The rotation quaternion of the local coordinate space.

        :return: `position` in local coordinates.
        """

        return QuaternionUtils.multiply_by_vector(q=QuaternionUtils.get_inverse(q=rotation), v=position - origin)

    @staticmethod
    def get_up_direction(q: np.array) -> np.array:
        """
        :param q: The rotation as a quaternion.

        :return: A directional vector corresponding to the "up" direction from the quaternion.
        """

        return QuaternionUtils.multiply_by_vector(q, QuaternionUtils.UP)

    @staticmethod
    def euler_angles_to_quaternion(euler: np.array) -> np.array:
        """
        Convert Euler angles to a quaternion.

        :param euler: The Euler angles vector.

        :return: The quaternion representation of the Euler angles.
        """

        roll = euler[0]
        pitch = euler[1]
        yaw = euler[2]
        cy = np.cos(yaw * 0.5)
        sy = np.sin(yaw * 0.5)
        cp = np.cos(pitch * 0.5)
        sp = np.sin(pitch * 0.5)
        cr = np.cos(roll * 0.5)
        sr = np.sin(roll * 0.5)

        w = cy * cp * cr + sy * sp * sr
        x = cy * cp * sr - sy * sp * cr
        y = sy * cp * sr + cy * sp * cr
        z = sy * cp * cr - cy * sp * sr
        return np.array([x, y, z, w])

    @staticmethod
    def quaternion_to_euler_angles(quaternion: np.array) -> np.array:
        """
        Convert a quaternion to Euler angles.

        :param quaternion: A quaternion as a nump array.

        :return: The Euler angles representation of the quaternion.
        """

        x = quaternion[0]
        y = quaternion[1]
        z = quaternion[2]
        w = quaternion[3]
        ysqr = y * y

        t0 = +2.0 * (w * x + y * z)
        t1 = +1.0 - 2.0 * (x * x + ysqr)
        ex = np.degrees(np.arctan2(t0, t1))

        t2 = +2.0 * (w * y - z * x)
        t2 = np.where(t2 > +1.0, +1.0, t2)

        t2 = np.where(t2 < -1.0, -1.0, t2)
        ey = np.degrees(np.arcsin(t2))

        t3 = +2.0 * (w * z + x * y)
        t4 = +1.0 - 2.0 * (ysqr + z * z)
        ez = np.degrees(np.arctan2(t3, t4))

        return np.array([ex, ey, ez])

    @staticmethod
    def get_y_angle(q1: np.array, q2: np.array) -> float:
        """
        Source: https://en.wikipedia.org/wiki/Conversion_between_quaternions_and_Euler_angles

        :param q1: The first quaternion.
        :param q2: The second quaternion.

        :return: The angle between the two quaternions in degrees around the y axis.
        """

        qd = QuaternionUtils.multiply(QuaternionUtils.get_conjugate(q1), q2)
        return np.rad2deg(2 * np.arcsin(np.clip(qd[1], -1, 1)))

    @staticmethod
    def is_left_of(origin: np.array, target: np.array, forward: np.array) -> bool:
        """
        :param origin: The origin position.
        :param target: The target position.
        :param forward: The forward directional vector.

        :return: True if `target` is to the left of `origin` by the `forward` vector; False if it's to the right.
        """

        # Get the heading.
        target_direction = target - origin
        # Normalize the heading.
        target_direction = target_direction / np.linalg.norm(target_direction)
        perpendicular: np.array = np.cross(forward, target_direction)
        direction = np.dot(perpendicular, QuaternionUtils.UP)
        return direction > 0



from typing import List, Union, Callable, Optional, Dict
from tdw.output_data import OutputData
from tdw.output_data import Keyboard as KBoard
from tdw.add_ons.add_on import AddOn


class Keyboard(AddOn):
    """
    Add keyboard controls to a TDW scene.
    """

    def __init__(self):
        super().__init__()
        self._press: Dict[str, Optional[Union[Callable, List[dict]]]] = dict()
        self._hold: Dict[str, Optional[Union[Callable, List[dict]]]] = dict()
        self._release: Dict[str, Optional[Union[Callable, List[dict]]]] = dict()

    def get_initialization_commands(self) -> List[dict]:
        return [{"$type": "send_keyboard",
                 "frequency": "always"}]

    def on_send(self, resp: List[bytes]) -> None:
        # Get keyboard input.
        for i in range(len(resp) - 1):
            r_id = OutputData.get_data_type_id(resp[i])
            if r_id == "keyb":
                keys = KBoard(resp[i])
                # Listen for events where the key was first pressed on the previous frame.
                for j in range(keys.get_num_pressed()):
                    self._do_event(k=keys.get_pressed(j), events=self._press)
                # Listen for keys currently held down.
                for j in range(keys.get_num_held()):
                    self._do_event(k=keys.get_held(j), events=self._hold)
                # Listen for keys that were released.
                for j in range(keys.get_num_released()):
                    self._do_event(k=keys.get_released(j), events=self._release)

    def listen(self, key: str, commands: Union[dict, List[dict]] = None, function: Callable = None,
               events: List[str] = None) -> None:
        """
        Listen for when a key is pressed and send commands.

        :param key: The keyboard key.
        :param commands: Commands to be sent when the key is pressed.
        :param function: Function to invoke when the key is pressed.
        :param events: Listen to these keyboard events for this `key`. Options: `"press"`, `"hold"`, `"release"`. If None, this defaults to `["press"]`.
        """

        response: Optional[Union[Callable, List[dict]]] = None
        if commands is not None:
            if isinstance(commands, dict):
                commands = [commands]
            response = commands
        elif function is not None:
            response = function
        if response is None:
            return
        if events is None:
            events = ["press"]

        # Subscribe to events.
        if "press" in events:
            self._press[key] = response
        if "hold" in events:
            self._hold[key] = response
        if "release" in events:
            self._release[key] = response

    def _do_event(self, k: str, events: dict) -> None:
        """
        Invoke an event if the key is in the events dictionary.

        :param k: The keyboard key as a string.
        :param events: The events dictionary.
        """

        if k in events:
            # Append commands to the next `communicate()` call.
            if isinstance(events[k], list):
                self.commands.extend(events[k])
            # Invoke a function.
            else:
                events[k]()



from abc import ABC, abstractmethod
from base64 import b64encode
from typing import List, Union, Dict
from pathlib import Path
import wave
import numpy as np
from overrides import final
from tdw.controller import Controller
from tdw.tdw_utils import TDWUtils
from tdw.add_ons.add_on import AddOn
from tdw.audio_constants import SAMPLE_RATE, CHANNELS


class AudioInitializerBase(AddOn, ABC):
    """
    Abstract base class for an audio initializer add-on.
    """

    def __init__(self, avatar_id: str = "a", framerate: int = 60):
        """
        :param avatar_id: The ID of the listening avatar.
        :param framerate: The target simulation framerate.
        """

        super().__init__()
        """:field
        The ID of the listening avatar.
        """
        self.avatar_id: str = avatar_id
        # The target framerate.
        self._target_framerate: int = framerate

    def get_initialization_commands(self) -> List[dict]:
        """
        This function gets called exactly once per add-on. To re-initialize, set `self.initialized = False`.

        :return: A list of commands that will initialize this add-on.
        """

        return [{"$type": "set_target_framerate",
                 "framerate": self._target_framerate},
                {"$type": self._get_sensor_command_name(),
                 "avatar_id": self.avatar_id}]

    @final
    def on_send(self, resp: List[bytes]) -> None:
        """
        This is called after commands are sent to the build and a response is received.

        Use this function to send commands to the build on the next frame, given the `resp` response.
        Any commands in the `self.commands` list will be sent on the next frame.

        :param resp: The response from the build.
        """

        return

    @final
    def play(self, path: Union[str, Path], position: Union[np.array, Dict[str, float]], audio_id: int = None,
             object_id: int = None) -> None:
        """
        Load a .wav file and prepare to send a command to the build to play the audio.
        The command will be sent on the next `Controller.communicate()` call.

        :param path: The path to a .wav file.
        :param position: The position of audio source. Can be a numpy array or x, y, z dictionary.
        :param audio_id: The unique ID of the audio source. If None, a random ID is generated.
        :param object_id: If not None, parent the audio source to this object.
        """

        if isinstance(path, Path):
            path = str(path.resolve())
        if isinstance(position, np.ndarray):
            position = TDWUtils.array_to_vector3(position)
        if audio_id is None:
            audio_id = Controller.get_unique_id()
        w = wave.open(path, 'rb')
        wav = w.readframes(w.getparams().nframes)
        self.commands.append({"$type": self._get_play_audio_command_name(),
                              "id": audio_id,
                              "position": position,
                              "num_frames": len(wav),
                              "num_channels": CHANNELS,
                              "frame_rate": SAMPLE_RATE,
                              "wav_data": str(b64encode(wav), 'ascii', 'ignore')})
        if object_id is not None:
            self.commands.append({"$type": "parent_audio_source_to_object",
                                  "object_id": object_id,
                                  "audio_id": audio_id})

    @abstractmethod
    def _get_sensor_command_name(self) -> str:
        """
        :return: The name of the command to add an audio sensor.
        """

        raise Exception()

    @abstractmethod
    def _get_play_audio_command_name(self) -> str:
        """
        :return: The name of the command to play audio.
        """

        raise Exception()



from pathlib import Path
import platform
from typing import List, Dict, Optional, Tuple, Union
from subprocess import call
import os
import shutil
from tdw.librarian import ModelRecord
import json
import pkg_resources
import distutils.dir_util
import distutils.file_util
from tdw.backend.platforms import S3_TO_UNITY, SYSTEM_TO_UNITY, UNITY_TO_SYSTEM
from tdw.asset_bundle_creator_base import AssetBundleCreatorBase
from tdw.add_ons.model_verifier.model_verifier import ModelVerifier
from tdw.controller import Controller


class AssetBundleCreator(AssetBundleCreatorBase):
    """
    Given a .fbx file or a .obj file, and (optionally) Materials and/or Textures folder adjacent to that file,
    create asset bundles for Windows, OS X, and Linux.

    Usage:

    ```python
    from tdw.asset_bundle_creator import AssetBundleCreator

    a = AssetBundleCreator()

    # Typically this is the only function you'll want to call.
    asset_bundle_paths, record_path = a.create_asset_bundle("cube.fbx", cleanup=True)
    ```

    For more information, see: `Documentation/misc_frontend/add_local_object.md`.
    """

    def __init__(self, quiet: bool = False, display: str = ":0", unity_editor_path: Union[Path, str] = None):
        """
        :param quiet: If true, don't print any messages to console.
        :param display: The display to launch Unity Editor on. Ignored if this isn't Linux.
        :param unity_editor_path: The path to the Unity Editor executable, for example `C:/Program Files/Unity/Hub/Editor/2020.3.24f1/Editor/Unity.exe`. If None, this script will try to find Unity Editor automatically.
        """

        super().__init__(quiet=quiet, display=display, unity_editor_path=unity_editor_path)

        system = platform.system()
        self._binaries: Dict[str, str] = dict()
        binary_path = f"binaries/{system}"

        # Cache the binaries.
        self._binaries["assimp"] = f"{binary_path}/assimp/assimp"
        self._binaries["meshconv"] = f"{binary_path}/meshconv/meshconv"
        self._binaries["vhacd"] = f"{binary_path}/vhacd/testVHACD"

        for binary in self._binaries:
            # Add the .exe suffix for Windows.
            if system == "Windows":
                self._binaries[binary] += ".exe"
            # Run chmod +x on everything.
            else:
                call(["chmod", "+x", pkg_resources.resource_filename(__name__, self._binaries[binary])])

    def create_asset_bundle(self, model_path: Union[Path, str], cleanup: bool, wnid: int = -1, wcategory: str = "", scale: float = 1) -> (List[Path], Path):
        """
        Create an asset bundle for each operating system. Typically, this is the only function you'll want to use.
        This function calls in sequence: `fbx_to_obj()`, `obj_to_wrl()`, `wrl_to_obj()`, `move_files_to_unity_project()`, `create_prefab()`, `prefab_to_asset_bundle()`, and `create_record()`.

        :param model_path: The path to the model.
        :param cleanup: If true, remove temporary files after usage.
        :param wnid: The WordNet ID.
        :param wcategory: The WordNet category.
        :param scale: The scale of the object.

        :return The paths to each asset bundle as Path objects (from pathlib) and the path to the metadata record file as a Path object (from pathlib).
        """

        model_path = self.get_model_path(model_path)

        model_name = model_path.stem

        # Create the asset bundles.
        obj_path, is_new = self.fbx_to_obj(model_path)
        wrl_path = self.obj_to_wrl(model_path)
        obj_colliders_path = self.wrl_to_obj(wrl_path, model_name)
        copied_file_paths = self.move_files_to_unity_project(obj_colliders_path, model_path)
        prefab_path, report_path = self.create_prefab(f"{model_name}_colliders.obj", model_name, model_path.suffix)
        asset_bundle_paths = self.prefab_to_asset_bundle(prefab_path, model_name)

        # Parse the URLs.
        urls = self.get_local_urls(asset_bundle_paths)

        # Create the metadata record.
        record_path = self.create_record(model_name, wnid, wcategory, scale, urls)

        # Remove the temporary files.
        if cleanup:
            paths = [wrl_path, obj_colliders_path, prefab_path, report_path]
            paths.extend(copied_file_paths)
            if is_new:
                paths.append(obj_path)

            assert model_path not in paths

            for path in paths:
                if not path.exists():
                    continue
                path.unlink()

            # Remove all materials.
            materials_directory = self.get_resources_directory().joinpath("Materials")
            if materials_directory.exists():
                shutil.rmtree(str(materials_directory.resolve()))

            if not self._quiet:
                print("Removed temporary files.")

        if not self._quiet:
            print("DONE!")

        return asset_bundle_paths, record_path

    @staticmethod
    def get_model_path(model_path: Union[Path, str]) -> Path:
        """
        Check if the model path is valid.

        :param model_path: The path to the model. Can be a Path object, or a string representing the absolute file path.

        :return The path as a Path object if there are no problems.
        """

        if isinstance(model_path, str):
            model_path = Path(model_path)
        assert model_path.exists(), f"Model path doesn't exist: {model_path}"
        assert model_path.is_file(), f"Model path isn't a file: {model_path}"
        assert model_path.suffix in [".fbx", ".obj", ".prefab"], "Model file must be .fbx, .obj, or .prefab"

        return model_path

    def get_assets_directory(self) -> Path:
        """
        :return The path to `<home>/asset_bundle_creator/Assets/`
        """

        assets_directory = self._project_path.joinpath("Assets")
        assert assets_directory.exists(), f"Assets directory not found: {assets_directory.resolve()}"

        return assets_directory

    def get_resources_directory(self) -> Path:
        """
        :return The path to `<home>/asset_bundle_creator/Assets/Resources`
        """

        resources_directory = self.get_assets_directory().joinpath("Resources")
        assert resources_directory.exists(), f"Resources directory not found: {resources_directory.resolve()}"

        return resources_directory

    @staticmethod
    def get_editor_path() -> Path:
        """
        :return The path to the Unity Editor executable.
        """

        return super().get_editor_path()

    def get_base_unity_call(self) -> List[str]:
        """
        :return The call to launch Unity Editor silently in batchmode, execute something, and then quit.
        """

        return super().get_base_unity_call()

    def get_unity_project(self) -> Path:
        """
        Build the asset_bundle_creator Unity project.

        :return The path to the asset_bundle_creator Unity project.
        """

        unity_project_path = self.get_project_path()

        # If the project already exists, stop.
        if unity_project_path.exists():
            return unity_project_path

        if not self._quiet:
            print(f"Creating: {unity_project_path.resolve()}")

        call([str(AssetBundleCreatorBase.get_editor_path().resolve()),
              "-createProject",
              str(unity_project_path.resolve()),
              "-quit",
              "-batchmode"], env=self._env)
        assert unity_project_path.exists(), unity_project_path.resolve()
        if not self._quiet:
            print(f"Created new Unity project: {str(unity_project_path.resolve())}")
        # Add the .unitypackage to the new project.
        package_name = "asset_bundle_creator.unitypackage"
        filepath = pkg_resources.resource_filename(__name__, package_name)
        assert Path(filepath).exists(), filepath
        # Import the package.
        call([str(AssetBundleCreatorBase.get_editor_path().resolve()),
              "-projectPath",
              str(unity_project_path.resolve()),
              "-importPackage",
              filepath,
              "-quit",
              "-batchmode"], env=self._env)
        if not self._quiet:
            print(f"Imported {package_name} into the new project.")

    @staticmethod
    def get_project_path() -> Path:
        """
        :return: The expected path of the Unity project.
        """

        return Path.home().joinpath("asset_bundle_creator")

    def fbx_to_obj(self, model_path: Path) -> Tuple[Path, bool]:
        """
        Convert a .fbx file to a .obj file with assimp

        :param model_path: The path to the model.

        :return The path to the new object, and True if it's a new file (False if it's the existing base file).
        """

        if model_path.suffix != ".fbx":
            assert model_path.suffix == ".obj"
            if not self._quiet:
                print("Model is already a .obj file. Skipping the conversion to .obj")
            return model_path, False

        if not self._quiet:
            print("Converting a .fbx file to .obj")

        # Create the .obj file.
        obj_filename = model_path.stem + ".obj"

        assimp = pkg_resources.resource_filename(__name__, self._binaries["assimp"])
        assert Path(assimp).exists(), assimp

        # Run assimp to create the .obj file.
        call([assimp,
              "export",
              str(model_path.resolve()),
              obj_filename],
             stdout=open(os.devnull, "wb"))

        obj_path = Path(obj_filename)
        assert obj_path.exists(), "Failed to convert .fbx to .obj"

        mtl_path = Path(obj_filename + ".mtl")
        if mtl_path.exists():
            mtl_path.unlink()
            if not self._quiet:
                print("Removed superfluous .obj.mtl file")

        return obj_path, True

    def obj_to_wrl(self, model_path: Path, vhacd_resolution: int = 8000000) -> Path:
        """
        Convert a .obj file to a .wrl file with testVHACD

        :param model_path: The path to the model.
        :param vhacd_resolution: The V-HACD voxel resolution. A higher number will create more accurate physics colliders, but it will take more time to initially create the asset bundle.

        :return The path to the .wrl file.
        """

        # Get the target .obj file.
        if model_path.suffix == ".obj":
            obj_path = str(model_path.resolve())
        else:
            obj_path = model_path.stem + ".obj"
        assert Path(obj_path).exists(), f"Target .obj doesn't exist: {obj_path}"

        wrl_filename = model_path.stem + ".wrl"
        if not self._quiet:
            print("Running V-HACD on a .obj file (this might take awhile).")

        # Run V-HACD.
        vhacd = pkg_resources.resource_filename(__name__, self._binaries["vhacd"])

        assert Path(vhacd).exists(), vhacd
        call([vhacd,
              "--input", obj_path,
              "--resolution", str(vhacd_resolution),
              "--output", wrl_filename],
             stdout=open(os.devnull, "wb"))

        assert Path(wrl_filename).exists(), "Failed to create .wrl file. " \
                                            "Check your original file to make sure that is 3D " \
                                            "(not just a flat plane) and isn't corrupted."

        if not self._quiet:
            print(f"Created: {wrl_filename}")

        # Remove the log, if any.
        if Path("log.txt").exists():
            os.remove("log.txt")

            if not self._quiet:
                print("Removed V-HACD log file.")

        # Remove the superfluous .obj file.
        if model_path.suffix != ".obj":
            os.remove(obj_path)

            if not self._quiet:
                print(f"Removed superfluous .obj file: {obj_path}")

        return Path(wrl_filename)

    def wrl_to_obj(self, wrl_filename: Path, model_name: str) -> Path:
        """
        Convert a .wrl file back into a .obj file with meshconv

        :param wrl_filename: The to the .wrl file.
        :param model_name: The name of the model (minus its file extension).

        :return The path to the .obj file.
        """

        assert wrl_filename.exists(), f"Missing .wrl file: {wrl_filename}"

        if not self._quiet:
            print("Converting .wrl to .obj")

        # Run meshconv.
        meshconv = pkg_resources.resource_filename(__name__, self._binaries["meshconv"])
        assert Path(meshconv).exists(), meshconv
        call([meshconv,
              str(wrl_filename.resolve()),
              "-c",
              "obj",
              "-o",
              f"{model_name}_colliders",
              "-sg"],
             stdout=open(os.devnull, "wb"))

        obj_path = Path(f"{model_name}_colliders.obj")
        assert obj_path.exists(), "Failed to create .obj file from .wrl file."

        if not self._quiet:
            print(f"Created {obj_path.resolve()} from .wrl file.")

        return obj_path

    def move_files_to_unity_project(self, obj_colliders: Optional[Path], model_path: Path, sub_directory: str = "") -> List[Path]:
        """
        Moves all required files to the Unity project.

        :param obj_colliders: The path to the colliders .obj file. May be None.
        :param sub_directory: Optional subdirectory to move the files to.
        :param model_path: The path to the model. Can be a Path object, or a string representing the absolute file path.

        :return A list of paths of files in the Unity project.
        """

        if obj_colliders is not None:
            assert obj_colliders.exists(), f"Colliders .obj file not found: {obj_colliders.resolve()}"

        resources_directory = self.get_resources_directory()

        if not self._quiet:
            print("Copying files into the Unity project.")

        mtl = model_path.parent.joinpath(Path(model_path.stem + ".mtl"))

        dests = []
        sources = [model_path, model_path.parent.joinpath("Materials"), model_path.parent.joinpath("Textures"), mtl]
        if obj_colliders is not None:
            sources.append(obj_colliders)

        # Open the mtl file and try to parse for textures
        if mtl.exists():
            mtl_txt = mtl.read_text()
            for mtl_line in mtl_txt.split("\n"):
                if ".jpg" in mtl_line:
                    tex_path = model_path.parent.joinpath(mtl_line.split(" ")[1])
                    assert tex_path.exists(), f"Missing texture: {tex_path}"
                    sources.append(tex_path)
        for src in sources:
            if src is None or not src.exists():
                continue
            if sub_directory == "":
                dest = resources_directory.joinpath(src.stem + src.suffix)
            else:
                dest = resources_directory.joinpath(sub_directory).joinpath(src.stem + src.suffix)
            # Copy to the destination.
            if src.is_file():
                if not dest.parent.exists():
                    Path(dest.parent).mkdir(parents=True)
                distutils.file_util.copy_file(str(src.resolve()), str(dest.resolve()))
            else:
                distutils.dir_util.copy_tree(str(src.resolve()), str(dest.resolve()))
            dests.append(dest)
        if not self._quiet:
            print("Copied files into the Unity project.")
        return dests

    def create_prefab(self, colliders: str, model_name: str, model_extension: str) -> Tuple[Path, Path]:
        """
        Create a prefab from the files existing in the Unity project folder.

        :param colliders: The colliders filename.
        :param model_name: The name of the model, minus its file extension.
        :param model_extension: The file extension of the model (e.g. ".obj").

        :return The path to the prefab and the path to the report (if any).
        """

        report = self.get_assets_directory().joinpath("report.txt")
        if report.exists():
            os.remove(str(report.resolve()))
            if not self._quiet:
                print("Removed old report.")

        # Create the prefab.
        if not self._quiet:
            print("Creating the prefab...")
        prefab_call = self._unity_call[:]
        prefab_call.extend(["-executeMethod",
                            "AssetBundleCreator.CreatePrefab",
                            "-filename=" + model_name,
                            "-modelname=" + model_name,
                            "-extension=" + model_extension,
                            "-colliders=" + colliders
                            ])
        call(prefab_call, env=self._env)
        prefab_path = self.get_resources_directory().joinpath(f"prefab/{model_name}.prefab")
        assert prefab_path.exists(), "Failed to create prefab."

        # Check the report.
        if report.exists():
            raise Exception(f"Created the prefab with errors: {report.read_text()}")

        if not self._quiet:
            print("Created the prefab.")

        return prefab_path, report

    def prefab_to_asset_bundle(self, prefab_path: Path, model_name: str, platforms: List[str] = None) -> List[Path]:
        """
        Given a .prefab, create asset bundles and write them to disk.

        :param prefab_path: The path to the .prefab file.
        :param model_name: The name of the model, minus its file extension.
        :param platforms: Platforms to build asset bundles for. Options: "windows", "osx", "linux". If None, build all.

        :return The paths to the asset bundles.
        """

        if platforms is None:
            platforms = ["windows", "osx", "linux"]

        assert prefab_path.exists(), f"Missing prefab: {prefab_path.resolve()}"

        if not self._quiet:
            print("Creating local asset bundles")
        asset_bundle_call = self._unity_call[:]
        platforms_call = ""
        for p in platforms:
            platforms_call += p + ","
        platforms_call = platforms_call[:-1]
        asset_bundle_call.extend(["-executeMethod",
                                  "AssetBundleCreator.BuildAssetBundle",
                                  "-modelname=" + model_name,
                                  "-platforms=" + platforms_call
                                  ])
        call(asset_bundle_call, env=self._env)
        new_asset_bundles_directory = self.get_assets_directory().joinpath("NewAssetBundles")
        new_asset_bundles_directory = new_asset_bundles_directory.joinpath(model_name)
        assert new_asset_bundles_directory.exists(), f"No asset bundles found: {new_asset_bundles_directory.resolve()}"

        paths = []

        # Iterate through all target platforms and build asset bundles.
        for platform_key in platforms:
            asset_bundle_platform = S3_TO_UNITY[platform_key]
            bundle_path = new_asset_bundles_directory.joinpath(asset_bundle_platform).joinpath(model_name)
            assert bundle_path.exists(), f"Missing asset bundle: {asset_bundle_platform}"
            paths.append(bundle_path)

        if not self._quiet:
            print("Created local asset bundles.")

        return paths

    @staticmethod
    def get_local_asset_bundle_path(model_name: str) -> Path:
        """
        :param model_name: The name of the model, minus its file extension.

        :return The expected path of the local asset bundle for this platform.
        """

        return Path.home().joinpath("asset_bundle_creator/Assets/NewAssetBundles").\
            joinpath(model_name + "/" + SYSTEM_TO_UNITY[platform.system()] + "/" + model_name)

    def create_record(self, model_name: str, wnid: int, wcategory: str, scale: float, urls: Dict[str, str], record: Optional[ModelRecord] = None, write_physics: bool = False) -> Path:
        """
        Create a local .json metadata record of the model.

        :param model_name: The name of the model.
        :param wnid: The WordNet ID.
        :param wcategory: The WordNet category.
        :param scale: The default scale of the object.
        :param urls: The finalized URLs (or local filepaths) of the assset bundles.
        :param record: A pre-written metadata record. If not None, it will override the other parameters.
        :param write_physics: If true, launch the build to write the physics quality. (This is optional).

        :return The path to the file with the metadata record.
        """

        # Write the record.
        if not self._quiet:
            print("Creating a record.")

        if record is None:
            record = ModelRecord()
            record.name = model_name
            record.wnid = f'n{wnid:08d}'
            record.wcategory = wcategory
            record.urls = urls
            record.scale = scale

        # Append asset bundle sizes.
        local_path = Path.home().joinpath("asset_bundle_creator/Assets/NewAssetBundles").joinpath(record.name)
        for os_dir in local_path.iterdir():
            if not os_dir.is_dir():
                continue
            asset_bundle_platform = UNITY_TO_SYSTEM[os_dir.stem]
            size = os_dir.joinpath(record.name).stat().st_size
            record.asset_bundle_sizes[asset_bundle_platform] = size

        # Assemble a dictionary of just the data that we don't need the Editor for.
        record_data = {"name": record.name,
                       "urls": record.urls,
                       "wnid": record.wnid,
                       "wcategory": record.wcategory,
                       "scale_factor": record.scale_factor,
                       "do_not_use": record.do_not_use,
                       "do_not_use_reason": record.do_not_use_reason,
                       "canonical_rotation": record.canonical_rotation,
                       "physics_quality": -1,
                       "asset_bundle_sizes": record.asset_bundle_sizes}

        # Serialize the record.
        record_data = json.dumps(record_data)
        # Remove the last } and replace it with , to keep serializing with Unity.
        record_data = record_data[:-1] + ","

        record_path = self.get_assets_directory().joinpath(model_name + ".json")
        record_path.write_text(record_data, encoding="utf-8")

        record_call = self._unity_call[:]
        record_call.extend(["-executeMethod",
                            "RecordCreator.WriteRecord",
                            "-modelname=" + model_name,
                            "-scale=" + str(scale)])
        call(record_call, env=self._env)

        # Test the record.
        try:
            json.loads(record_path.read_text(encoding="utf-8"))
        except json.JSONDecodeError:
            raise Exception("Failed to deserialize: " + record_path.read_text(encoding="utf-8"))

        if not self._quiet:
            print("Wrote the record data to the disk.")

        if write_physics:
            self.write_physics_quality(record_path=record_path,
                                       asset_bundle_path=self.get_local_asset_bundle_path(model_name).resolve())

        return record_path

    @staticmethod
    def get_local_urls(asset_bundle_paths: List[Path]) -> Dict[str, str]:
        """
        Generate a dictionary of local URLs from the asset bundle paths.

        :param asset_bundle_paths: The asset bundle paths.

        :return A dictionary. Key = OS, Value = Path to the local file.
        """

        urls: Dict[str, str] = {}
        for ap in asset_bundle_paths:
            ap = "file:///" + str(ap.resolve()).replace("\\", "/")
            if "StandaloneWindows64" in ap:
                urls.update({"Windows": ap})
            elif "StandaloneOSX" in ap:
                urls.update({"Darwin": ap})
            elif "StandaloneLinux64" in ap:
                urls.update({"Linux": ap})
        return urls

    @staticmethod
    def write_physics_quality(record_path: Path, asset_bundle_path: Path) -> None:
        """
        Append the physics quality data to the temporary record file.
        This is an optional record field that records the percentage of the model encapsualted by colliders.

        :param record_path: The path to the temporary record file.
        :param asset_bundle_path: The URL to the local asset bundle.
        """

        c = Controller()
        v = ModelVerifier()
        r: ModelRecord = ModelRecord(json.loads(Path(record_path).read_text(encoding="utf-8")))
        r.urls[platform.system()] = f"file:///{str(asset_bundle_path)}"
        v.set_tests(name=r.name, source=r, model_report=False, missing_materials=False, physics_quality=True)
        c.add_ons.append(v)
        c.communicate([])
        while not v.done:
            c.communicate([])
        c.communicate({"$type": "terminate"})
        c.socket.close()
        # Write the physics quality.
        r.physics_quality = float(v.reports[0])
        record_path.write_text(json.dumps(r.__dict__), encoding="utf-8")

    def validate(self, record_path: Path, asset_bundle_path: Path) -> Tuple[bool, str]:
        """
        Validate the asset bundle.

        :param record_path: The path to the temporary record file.
        :param asset_bundle_path: The URL to the local asset bundle.

        :return True if there aren't problems, and a string output report.
        """

        if not self._quiet:
            print("Validating asset bundle...")

        c = Controller()
        v = ModelVerifier()
        r: ModelRecord = ModelRecord(json.loads(Path(record_path).read_text(encoding="utf-8")))
        r.urls[platform.system()] = f"file:///{str(asset_bundle_path)}"
        v.set_tests(name=r.name, source=r, model_report=True, missing_materials=True, physics_quality=False)
        c.add_ons.append(v)
        c.communicate([])
        while not v.done:
            c.communicate([])
        c.communicate({"$type": "terminate"})
        c.socket.close()
        if len(v.reports) != 0:
            output = "There are problems with the asset bundle!"
            for problem in v.reports:
                output += "\n\t" + problem
            return False, output
        if not self._quiet:
            print("OK!")
        return True, ""

    def create_many_asset_bundles(self, library_path: str, cleanup: bool = True, vhacd_resolution: int = 8000000) -> None:
        """
        Create asset bundles in batch from .obj or .fbx files already in: asset_bundle_creator/Assets/Resources/models
        This function will create collider .obj files if there aren't any already
        (it will look for a file named <model>_colliders.obj).

        :param library_path: The path to the library file.
        :param cleanup: If true, remove all temp files when done.
        :param vhacd_resolution: The V-HACD voxel resolution. A higher number will create more accurate physics colliders, but it will take more time to initially create the asset bundle.
        """

        # Create for each model that doesn't have colliders.
        models_dir = self.get_resources_directory().joinpath("models")
        for ext in [".obj", ".fbx"]:
            for f in models_dir.rglob(f"*{ext}"):
                if f.stem.endswith("_colliders"):
                    continue
                colliders_path = f.parent.joinpath(f"{f.stem}_colliders.obj")
                # Don't regenerate a colliders file if one already exists.
                if colliders_path.exists():
                    continue
                # Don't recreate models that already have prefabs.
                prefab_path = self.get_resources_directory().joinpath(f"prefab/{f.stem}.prefab")
                if prefab_path.exists():
                    continue
                # Create the colliders.
                wrl_path = self.obj_to_wrl(f, vhacd_resolution=vhacd_resolution)
                # Move the collider .obj to the correct directory.
                wrl_to_obj_path = self.wrl_to_obj(wrl_path, f.stem)
                distutils.file_util.move_file(str(wrl_to_obj_path.resolve()), str(colliders_path.resolve()))
                # Remove the .wrl file.
                wrl_path.unlink()

        # Create the asset bundles.
        record_call = self._unity_call[:]
        record_call.extend(["-executeMethod",
                            "AssetBundleCreator.CreateManyAssetBundles",
                            "-library=" + library_path,
                            "-internal_materials"])
        call(record_call)

        if cleanup:
            self.cleanup()

    def cleanup(self) -> None:
        """
        Delete all files from `~/asset_bundle_creator` with these extensions: .obj, .fbx, .mtl, .mat, .jpg, .prefab
        """

        root_dir = self.get_assets_directory()
        # Remove assets used to create asset bundles.
        for ext in [".obj", ".fbx", ".mtl", ".mat", ".jpg", ".prefab"]:
            for f in root_dir.rglob(f"*{ext}"):
                f.unlink()
        # Remove asset bundle junk.
        distutils.dir_util.remove_tree(str(self.get_assets_directory().joinpath("NewAssetBundles").resolve()))
        # Remove models junk.
        models_dir = self.get_resources_directory().joinpath("models")
        for d in models_dir.iterdir():
            if "placeholder" in d.stem:
                continue
            if d.is_file():
                d.unlink()
            else:
                distutils.dir_util.remove_tree(str(d.resolve()))



from secrets import token_urlsafe
from abc import ABC, abstractmethod
from typing import List, Dict, Optional
from overrides import final
from tdw.add_ons.add_on import AddOn


class ThirdPersonCameraBase(AddOn, ABC):
    """
    An abstract base class for third-person camera controller add-ons.
    """

    # The render order. Third person cameras will always render "on top" of any other cameras.
    __RENDER_ORDER: int = 100

    def __init__(self, avatar_id: str = None, position: Dict[str, float] = None, rotation: Dict[str, float] = None,
                 field_of_view: int = None):
        """
        :param avatar_id: The ID of the avatar (camera). If None, a random ID is generated.
        :param position: The initial position of the camera. If None, defaults to `{"x": 0, "y": 0, "z": 0}`.
        :param rotation: The initial rotation of the camera. Can be Euler angles (keys are `(x, y, z)`) or a quaternion (keys are `(x, y, z, w)`). If None, defaults to `{"x": 0, "y": 0, "z": 0}`.
        :param field_of_view: If not None, set the field of view.
        """

        super().__init__()
        # Set a random avatar ID.
        if avatar_id is None:
            """:field
            The ID of the avatar that (this camera).
            """
            self.avatar_id: str = token_urlsafe(4)
        else:
            self.avatar_id: str = avatar_id
        self._render_order = ThirdPersonCameraBase.__RENDER_ORDER
        ThirdPersonCameraBase.__RENDER_ORDER += 1
        """:field
        The position of the camera. If None, defaults to `{"x": 0, "y": 0, "z": 0}`.
        """
        self.position: Optional[Dict[str, float]] = position
        # The initial rotation.
        self._rotation: Optional[Dict[str, float]] = rotation
        # The field of view.
        self._field_of_view: float = field_of_view

    @abstractmethod
    def on_send(self, resp: List[bytes]) -> None:
        """
        This is called after commands are sent to the build and a response is received.

        Use this function to send commands to the build on the next frame, given the `resp` response.
        Any commands in the `self.commands` list will be sent on the next frame.

        :param resp: The response from the build.
        """

        raise Exception()

    def get_initialization_commands(self) -> List[dict]:
        """
        This function gets called exactly once per add-on. To re-initialize, set `self.initialized = False`.

        :return: A list of commands that will initialize this add-on.
        """

        commands = [{"$type": "create_avatar",
                     "type": self._get_avatar_type(),
                     "id": self.avatar_id},
                    {"$type": "set_pass_masks",
                     "pass_masks": ["_img"],
                     "avatar_id": self.avatar_id},
                    {"$type": "set_render_order",
                     "render_order": self._render_order,
                     "avatar_id": self.avatar_id}]
        # Set the initial position.
        if self.position is not None:
            commands.append({"$type": "teleport_avatar_to",
                             "position": self.position,
                             "avatar_id": self.avatar_id})
        # Set the initial rotation.
        if self._rotation is not None:
            if isinstance(self._rotation, dict):
                if "w" in self._rotation:
                    commands.append({"$type": "rotate_sensor_container_to",
                                     "rotation": self._rotation,
                                     "avatar_id": self.avatar_id})
                else:
                    for q, axis in zip(["x", "y", "z"], ["pitch", "yaw", "roll"]):
                        commands.append({"$type": "rotate_sensor_container_by",
                                         "axis": axis,
                                         "angle": self._rotation[q],
                                         "avatar_id": self.avatar_id})
        # Set the field of view.
        if self._field_of_view is not None:
            commands.append({"$type": "set_field_of_view",
                             "field_of_view": self._field_of_view,
                             "avatar_id": self.avatar_id})
        return commands

    @final
    def before_send(self, commands: List[dict]) -> None:
        """
        This is called before sending commands to the build. By default, this function doesn't do anything.

        :param commands: The commands that are about to be sent to the build.
        """

        pass

    def _get_avatar_type(self) -> str:
        """
        :return: The type of avatar.
        """

        return "A_Img_Caps_Kinematic"



from typing import List, Dict, Union
import numpy as np
from overrides import final
from ikpy.chain import Chain
from ikpy.link import OriginLink, URDFLink, Link
from tdw.tdw_utils import TDWUtils
from tdw.quaternion_utils import QuaternionUtils
from tdw.add_ons.robot import Robot
from tdw.librarian import RobotLibrarian, RobotRecord


class RobotArm(Robot):
    """
    A robot with a single arm.
    This class includes an inverse kinematic (IK) solver that allows the robot to reach for a target position.
    """

    def __init__(self, name: str, robot_id: int = 0, position: Dict[str, float] = None, rotation: Dict[str, float] = None,
                 source: Union[RobotLibrarian, RobotRecord] = None):
        """
        :param name: The name of the robot.
        :param robot_id: The ID of the robot.
        :param position: The position of the robot. If None, defaults to `{"x": 0, "y": 0, "z": 0}`.
        :param rotation: The rotation of the robot in Euler angles (degrees). If None, defaults to `{"x": 0, "y": 0, "z": 0}`.
        :param source: The source file of the robot. If None: The source will be the URL of the robot record in TDW's built-in [`RobotLibrarian`](../librarian/robot_librarian.md). If `RobotRecord`: the source is the URL in the record. If `RobotLibrarian`: The source is the record in the provided `RobotLibrarian` that matches `name`.
        """

        super().__init__(name=name, robot_id=robot_id, position=position, rotation=rotation, source=source)
        assert self._record is not None, "Record is None."
        assert len(self._record.ik) > 0, "Record doesn't have IK data."
        # A list of joint names in the order that they appear in the IK chain.
        self._joint_order: List[str] = list()
        # Convert the record data into joint links.
        links: List[Link] = [OriginLink()]
        for link in self._record.ik[0]:
            self._joint_order.append(link["name"])
            links.append(URDFLink(name=link["name"],
                                  translation_vector=np.array(link["translation_vector"]),
                                  orientation=np.array(link["orientation"]),
                                  rotation=None if link["rotation"] is None else np.array(link["rotation"]),
                                  bounds=link["bounds"]))
        # Set robot arm IK chain.
        self._chain: Chain = Chain(name=name, links=links)

    def reach_for(self, target: Union[Dict[str, float], np.array]) -> None:
        """
        Start to reach for a target position.

        :param target: The target position. Can be a dictionary or a numpy array.
        """

        angles = self._get_ik_angles(target=target)
        # Convert the IK solution to degrees. Remove the origin link.
        angles = [float(np.rad2deg(angle)) for angle in angles[1:]]
        # Convert the angles to a dictionary of joint targets.
        targets = dict()
        for joint_name, angle in zip(self._joint_order, angles):
            targets[self.static.joint_ids_by_name[joint_name]] = angle
        self.set_joint_targets(targets=targets)

    def set_joint_targets(self, targets: Dict[int, Union[float, Dict[str, float]]]) -> None:
        """
        Set target angles or positions for a dictionary of joints.

        :param targets: A dictionary of joint targets. Key = The ID of the joint. Value = the targets. For spherical joints, this must be a Vector3 dictionary, for example `{"x": 40, "y": 0, "z": 0}` (angles in degrees). For revolute joints, this must be a float (an angle in degrees). For prismatic joints, this must be a float (a distance in meters).
        """

        super().set_joint_targets(targets=targets)

    def add_joint_forces(self, forces: Dict[int, Union[float, Dict[str, float]]]) -> None:
        """
        Add torques and forces to a dictionary of joints.

        :param forces: A dictionary of joint forces. Key = The ID of the joint. Value = the targets. For spherical joints, this must be a Vector3 dictionary, for example `{"x": 40, "y": 0, "z": 0}` (torques in Newtons). For revolute joints, this must be a float (a torque in Newtons). For prismatic joints, this must be a float (a force in Newtons).
        """

        super().add_joint_forces(forces=forces)

    def stop_joints(self, joint_ids: List[int] = None) -> None:
        """
        Stop the joints at their current angles or positions.

        :param joint_ids: A list of joint IDs. If None, stop all joints.
        """

        super().stop_joints(joint_ids=joint_ids)

    @final
    def _get_ik_angles(self, target: Union[Dict[str, float], np.array]) -> List[float]:
        """
        :param target: The target position to reach for.

        :return: A list of angles of an IK solution in radians.
        """

        # Get the current angles of the joints.
        initial_angles = [0]
        for joint_name in self._joint_order:
            initial_angles.append(self.dynamic.joints[self.static.joint_ids_by_name[joint_name]].angles[0])
        initial_angles = np.radians(initial_angles)
        if isinstance(target, dict):
            target = TDWUtils.vector3_to_array(target)
        # Convert the worldspace position to a relative position.
        relative_target = self._absolute_to_relative(target=target)
        # Get the IK solution.
        return self._chain.inverse_kinematics(target_position=relative_target, initial_position=initial_angles)

    @final
    def _absolute_to_relative(self, target: np.array) -> np.array:
        """
        :param target: The target position.
        :return: The target position in relative coordinates.
        """

        return QuaternionUtils.world_to_local_vector(position=target,
                                                     origin=self.dynamic.transform.position,
                                                     rotation=self.dynamic.transform.rotation)
